"""Formatter producing CSV data."""

import sys
import csv
from collections import OrderedDict
import pvcheck.formatter
import pvcheck.executor
from pvcheck.i18n import translate as _

# TO BE DEFINED
# - wrong lines
# - return_code/code/section status
# - difference/similarity


class CSVFormatter(pvcheck.formatter.Formatter):
    _RESULT_TABLE = {
        pvcheck.executor.ER_OK: "0",
        pvcheck.executor.ER_TIMEOUT: "1",
        pvcheck.executor.ER_SEGFAULT:
        "2",
        pvcheck.executor.ER_ERROR:
        "3",
        pvcheck.executor.ER_NOTFILE:
        "4"
    }

    def __init__(self, destination=sys.stdout):
        self._dest = destination
        self._obj = None
        self._tests = []
        
    def _proc_args(self, args):
        return [(a if a is not pvcheck.executor.ARG_TMPFILE
                 else "<temp.file>")
                for a in args]

    def begin_session(self):
        self._tests = []

    def _header_builder(self):
        """Build the header.

        If there is only a test omits the header 'TEST'.

        """
        header = ["TEST"]
        header.append(_("CODE"))
        for test in self._tests:
            section_names = list(test["sections"].keys())
            for name in section_names:
                if name not in header:
                    header.append(name)
        return header

    def _row_builder(self, test, header):
        """Build a row.

        If there is only a test omits the test's name.

        """
        row = [test["title"]]
        for element in header:
            if element == _("CODE"):
                row.append(test["status"])
            elif element != "TEST":
                try:
                    row.append(test["sections"][element]["equality"])
                except KeyError:
                    row.append("")
        return row

    def _statistics_row_builder(self, header):
        """Build a row containing the arithmetic mean of equality for each section."""
        row = [_("TOTAL"), ""]
        for head in header:
            if head not in ("TEST", _("CODE")):
                values = []
                for test in self._tests:
                    try:
                        if test["sections"][head]["equality"] != 'MISS':
                            values.append(float(test["sections"][head]["equality"]))
                        else:
                            values.append(0)
                    except KeyError:
                        values.append(None)
                try:
                    # takes into account the sections that actually exist
                    values_2 = [v for v in values if v is not None]
                    row.append('%.2f' % (sum(values_2)/len(values_2)))
                except ZeroDivisionError:
                    row.append('%.2f' % 0)
        return row

    def end_session(self):
        header = self._header_builder()
        fp = csv.writer(self._dest)
        fp.writerow(header)
        for test in self._tests:
            row = self._row_builder(test, header)
            fp.writerow(row)
        fp.writerow(self._statistics_row_builder(header))

    def begin_test(self, description, cmdline_args, input, tempfile):
        t = OrderedDict([
            ("title", description)
        ])
        self._tests.append(t)

    def execution_result(self, cmdline_args, execution_result, test):
        t = self._tests[-1]
        self._sections = OrderedDict()
        t["status"] = self._RESULT_TABLE[execution_result.result]
        if execution_result.result != pvcheck.executor.ER_OK:
            for s in test.sections(exclude_special=True):
                self._sections[s.tag] = OrderedDict([("equality", "0")])

        t["sections"] = self._sections

    def comparison_result(self, expected, got, diffs, matches):
        if len(diffs) != 0:
            percent_correct = '%.2f' % (((len(diffs) - sum(diffs)) * 100)/len(diffs))
        else:
            percent_correct = '100.00'
        s = OrderedDict([
            ("equality", percent_correct)
        ])
        self._sections[expected.tag] = s

    def missing_section(self, expected):
        s = OrderedDict([("equality", "MISS")])
        self._sections[expected.tag] = s



"""Formatter producing HTML data"""
from pvcheck.jsonformatter import JSONFormatter
import pvcheck.i18n


_ = pvcheck.i18n.translate
_trans_dic = {"\n": "<br>", "–": "&ndash;", "—": "&mdash;", "&": "&amp;", ">": "&gt;", "<": "&lt;"}
_trantab = str.maketrans(_trans_dic)

section_summary = {}
total_summary = {"ok": 0, "warning": 0, "error": 0}


class HTMLFormatter(JSONFormatter):

    def end_session(self):

        header = self._tests_table_header_builder()
        for element in header:
            if element != "TEST":
                section_summary[element] = {"ok": 0, "warning": 0, "error": 0}

        self.print_html_header()
        self.print_tests_table()
        self.print_tests_information()
        self.print_summary_table()
        print("    </body>")
        print('</html>')

    @staticmethod
    def print_html_header():
        """Print header and style."""
        print("""<!DOCTYPE html>
<html>
    <head>
        <meta charset="utf-8">
        <title>PvCheck Result</title>
        <style rel="stylesheet" type="text/css">

            html {
              font-family: sans-serif;
            }
            
            h1 {
             overflow: hidden;
             text-align: center;
            }
            h1:before,
            h1:after {
             background-color: #333;
             content: "";
             display: inline-block;
             height: 1px;
             position: relative;
             vertical-align: middle;
             width: 50%;
            }
            h1:before {
             right: 0.5em;
             margin-left: -50%;
            }
            h1:after {
             left: 0.5em;
             margin-right: -50%;
            }
            
            table {
              border-collapse: collapse;
              border: 2px solid rgb(200,200,200);
              letter-spacing: 1px;
              font-size: 0.8rem;
            }

            td, th {
              border: 1px solid rgb(190,190,190);
              padding: 10px 20px;
            }

            th {
              background-color: rgb(235,235,235);
            }

            td {
              text-align: center;
            }

            tr:nth-child(even) td {
              background-color: rgb(250,250,250);
            }

            tr:nth-child(odd) td {
              background-color: rgb(245,245,245);
            }

            caption {
              padding: 10px;
            }
        </style>
    </head>
    <body>""")
        print("""       <h1>PvCheck</h1>
        <ul>
            <li><a href="#summary">{}</a></li>
            <li><a href="#info">info</a></li>
        </ul>""".format(_("summary")))

    def print_tests_table(self):
        """Print a table containing tests' results."""
        print('        <h2 align="center">{}</h2>'.format(_("Test Result")))
        print('        <table align="center">')
        print('            <tr>')

        tests_table_header = self._tests_table_header_builder()

        self._print_tests_table_header(tests_table_header)
        self._print_tests_table_rows(tests_table_header)

        print("        </table>")
        print("        <br>")

    def _tests_table_header_builder(self):
        if len(self._tests) > 1:
            header = ["TEST"]
        else:
            header = []
        for test in self._tests:
            section_names = list(test["sections"].keys())
            for name in section_names:
                if name not in header:
                    header.append(name)
        return header

    @staticmethod
    def _print_tests_table_header(tests_table_header):
        for element in tests_table_header:
            print("                <th>{}</th>".format(element.translate(_trantab)))
        print("            </tr>")

    def _print_tests_table_rows(self, tests_table_header):
        if len(self._tests) > 1:
            print_test_name = True
            first_element_index = 1
        else:
            print_test_name = False
            first_element_index = 0
        for test in self._tests:
            row = self._test_table_row_builder(test, tests_table_header)
            self._print_test_table_row(print_test_name, first_element_index, row)

    def _test_table_row_builder(self, test, header):
        if len(self._tests) > 1:
            row = [test["title"]]
        else:
            row = []
        for element in header:
            if element != "TEST":
                try:
                    row.append(test["sections"][element]["section status"])
                except KeyError:
                    row.append("")
        return row

    def _print_test_table_row(self, print_test_name, first_element_index, row):
        print("            <tr>")
        if print_test_name:
            print('                <td><a href="#{}">{}</a></td>'.format(row[0].translate(_trantab),
                                                                         row[0].translate(_trantab)))
        for element in row[first_element_index:]:
            self._print_section_status(element)
        print("            </tr>")

    @staticmethod
    def _print_section_status(section):
        if section == "ok":
            color = "green"
        elif section == "error":
            color = "red"
        elif section == "exec_error":
            section = "execution failed"
            color = "red"
        elif section == "missing":
            color = "orange"
        else:
            color = "black"
        print('                <td><font color="{}">{}</font></td>'.format(color, section.translate(_trantab)))

    def print_tests_information(self):
        """Print a section for each test containing results' information."""
        print('        <h2><a name="info">Info</a></h2>')
        if self._tests[0]["error_message"] != "ok":
            self._print_error_message()
            exit(0)
        header = self._tests_table_header_builder()
        for test in self._tests:
            self._print_test_information(test, header)
        print("        <hr>")

    def _print_error_message(self):
        print("        <hr>")
        print("        <p><font color='red'>{}</font></p>".format(self._tests[0]["error_message"]))
        print("        <hr>")
        print("    </body>")
        print('</html>')

    def _print_test_information(self, test, sections):
        print("        <hr>")
        if len(self._tests) > 1:
            self._print_test_name(test)
        self._print_command_line(test)
        if len(test["input_text"]) > 0:
            self._print_input_text(test)
        if test["input_file_name"] is not None:
            self._print_input_file_name(test)
        for section in sections:
            if section != "TEST":
                self._print_section_status_message(test, section)
        print('        </p>')

    @staticmethod
    def _print_test_name(test):
        print('        <p><a name="{}"><b>TEST:</b> {}</a><br>'.format(test["title"].translate(_trantab),
                                                                       test["title"].translate(_trantab)))

    @staticmethod
    def _print_command_line(test):
        command_line = ""
        for element in test["command_line"]:
            command_line += " " + element
        print('            <b>{}:</b> {}<br>'.format(_("COMMAND LINE"), command_line.translate(_trantab)))

    @staticmethod
    def _print_input_text(test):
        print('            <b>INPUT:</b> {}<br>'.format(test["input_text"].translate(_trantab)))

    @staticmethod
    def _print_input_file_name(test):
        if test["input_file_name"] == "<temp.file>":
            input_file_name = _("TEMPORARY FILE")
        else:
            input_file_name = test["input_file_name"]

        print('            <b>{}:</b><br> {}<br>'.format(input_file_name.translate(_trantab),
                                                         test["file_text"].translate(_trantab)))

    def _print_section_status_message(self, test, section):
        if test["sections"][section]["section status"] == "ok":
            section_summary[section]["ok"] += 1
            total_summary["ok"] += 1
            self._print_section_ok_message(section)
        elif test["sections"][section]["section status"] == "error":
            section_summary[section]["error"] += 1
            total_summary["error"] += 1
            for wrong_line in test["sections"][section]['wrong_lines']:
                self._print_section_error_message(wrong_line, section)
        elif test["sections"][section]["section status"] == "exec_error":
            section_summary[section]["error"] += 1
            total_summary["error"] += 1
            self._print_section_exec_error_message(section)
        else:
            section_summary[section]["warning"] += 1
            total_summary["warning"] += 1
            self._print_section_warning_message(section)

    @staticmethod
    def _print_section_ok_message(section):
        msg = "OK"
        color = "green"
        print('            <b><font color="{}">{}: </b>{}</font><br>'.format(color, section, msg))

    @staticmethod
    def _print_section_error_message(wrong_line, section):
        if wrong_line[2] is None:
            msg = _("unexpected line '%s'") % (wrong_line[0] + 1)
        elif wrong_line[1] is None:
            msg = _("missing line (expected '%s')") % (wrong_line[2])
        else:
            out_string = pvcheck.formatter.handle_non_printable_chars(wrong_line[1])
            out_string = out_string.translate(_trantab)
            msg = _("line %d is wrong  (expected '%s', got '%s')") % (wrong_line[0] + 1, wrong_line[2],
                                                                      out_string)
        color = "red"
        print('            <b><font color="{}">{}: </b>{}</font><br>'.format(color, section, msg))

    @staticmethod
    def _print_section_exec_error_message(section):
        msg = _("execution failed")
        color = "red"
        print('            <b><font color="{}">{}: </b>{}</font><br>'.format(color, section, msg))

    @staticmethod
    def _print_section_warning_message(section):
        msg = _("missing section")
        color = "orange"
        print('            <b><font color="{}">{}: </b>{}</font><br>'.format(color, section, msg))

    def print_summary_table(self):
        """Print a test summary table."""
        self._print_summary_table_header()
        header = self._tests_table_header_builder()
        for section in header:
            if section != "TEST":
                self._print_section_summary_row(section)
        self._print_total_summary_row()
        print("        </table>")

    @staticmethod
    def _print_summary_table_header():
        print("""        <a name="summary"><h2 align="center">{}</h2>
        <table align="center">
            <tr>
                <th>&nbsp;</th>
                <th><font color= "green">{}</font></th>
                <th><font color = "orange">{}</font></th>
                <th><font color = "red">{}</font></th>
            </tr>""".format(_("Summary"), _("Successes"), _("Warnings"), _("Errors")))

    @staticmethod
    def _print_section_summary_row(section):
        print("            <tr>")
        print('                <td>{}</td>'.format(section))
        print('                <td>{}</td>'.format(section_summary[section]["ok"]))
        print('                <td>{}</td>'.format(section_summary[section]["warning"]))
        print('                <td>{}</td>'.format(section_summary[section]["error"]))
        print("            </tr>")

    @staticmethod
    def _print_total_summary_row():
        print("            <tr>")
        print('                <td>{}</td>'.format(_("TOTAL")))
        print('                <td>{}</td>'.format(total_summary["ok"]))
        print('                <td>{}</td>'.format(total_summary["warning"]))
        print('                <td>{}</td>'.format(total_summary["error"]))
        print("            </tr>")



#!/usr/bin/env python3
# loaded here to change the default translation backend of gettext with ours
import pvcheck.i18n
import gettext
# this changes the default translation backend - must be done before any
# import that uses gettext (e.g. argparser)
gettext.gettext = pvcheck.i18n.translate

import argparse
import sys

from pvcheck.argparser import ArgParser
import os
import math
import pvcheck.pvcheck
import pvcheck.parser
import pvcheck.testdata
import pvcheck.formatter
import pvcheck.jsonformatter
import pvcheck.csvformatter
import pvcheck.htmlformatter
import pvcheck.interactiveformatter
import pvcheck.executor
import pvcheck.valgrind
import pvcheck.i18n
import pvcheck.exporter


_ = pvcheck.i18n.translate
_DEFAULT_LOG_FILE = os.path.expanduser("~/.pvcheck.log")


def parse_options():
    """Parse the command line."""
    argparser = _initialized_argparser()
    argparser.set_default_subparser("run")
    args = argparser.parse_args()

    verbosity = args.verbosity
    timeout = args.timeout
    config = args.config
    color = args.color
    color = (color == "YES" or (color == "AUTO" and sys.stdout.isatty()))
    format = args.format
    logfile = args.log
    valgrind = args.valgrind
    run = args.test
    export = args.test_number
    list = args.info
    test_file = args.file
    program = args.program
    program_arguments = args.program_arguments
    maxerrors = args.errors
    output_limit = args.output_limit

    args = dict(test_file=test_file, program=program, program_arguments=program_arguments
                )
    opts = dict(config=config, verbosity=verbosity, timeout=timeout,
                maxerrors=maxerrors, color=color, valgrind=valgrind,
                format=format, logfile=logfile, list=list, run=run, export=export, output_limit=output_limit)
    return args, opts


def _initialized_argparser():
    argparser = ArgParser(description=_("Run tests to verify the correctness of a program."))

    subparsers = argparser.add_subparsers(help=_('[run|info|export] --help for command help (default=run)'))

    # create the parser for the "run" command
    parser_run = subparsers.add_parser('run', help=_('test a program.'))
    a = parser_run.add_argument
    a("program", help=_("program to be tested."))
    a("program_arguments", help=_("any arguments of the program to be tested."), nargs='*')

    a("-f", "--file", help=_("file containing the tests to be performed (default pvcheck.test).")
                            , default="pvcheck.test")
    a("-T", "--test", help=_("run only the selected test."), nargs="?", type=int)
    a("-t", "--timeout", help=_("set how many seconds it should be waited for the termination "
                            "of the program.  The default is 10 seconds."), nargs='?', const=10, default=10,
                            type=check_float_non_negative)
    a("-e", "--errors", help=_("reports up to N errors per section (default 4)."), nargs='?',
                             const=4, default=4, type=check_int_greater_than_one)
    a("-v", "--verbosity", help=_("set the verbosity level, where the level must be an integer "
                            "between 0 (minimum) and 4 (maximum). The default value is 3."), nargs='?', const=3,
                            default=3, type=int, choices=range(0, 5))
    a("-V", "--valgrind", help=_("use Valgrind (if installed) to check memory usage."),
                            action='store_true')
    a("-l", "--log", help=_("specify the name of the file used for logging.  The default is "
                            "~/.pvcheck.log."), nargs='?', const=_DEFAULT_LOG_FILE, default=_DEFAULT_LOG_FILE)
    a("-L", "--output_limit", help=_("cut the output of the program to a maximum of L lines.  "
                            "The default is 10000."), nargs='?', const=10000, default=10000,
                            type=check_int_non_negative)
    a("-c", "--config", help=_("uses the specified configuration file."), nargs='?', const='',
                            default='')
    a("-F", "--format", help=_("select the output type."), default='interactive',
      choices=('interactive', 'text', 'json', 'csv', 'html'))
    a("-C", "--color", help=_("enable or disable colored output (default AUTO)."), nargs='?',
                               const='AUTO', default='AUTO', choices=('YES', 'NO', 'AUTO'))

    parser_run.set_defaults(test_number=None, info=False)

    # create the parser for the "info" command
    parser_info = subparsers.add_parser('info', help=_("list all the available tests."))

    parser_info.add_argument("file", help=_("file containing the tests to be performed."))
    parser_info.set_defaults(config='', timeout=10, verbosity=3, errors=4, color='AUTO', valgrind=False,
                             format='text', log=_DEFAULT_LOG_FILE, test=None, program=None, program_arguments=None,
                             test_number=None, info=True, output_limit=10000)

    # create the parser for the "export" command
    parser_export = subparsers.add_parser('export', help=_("export in a file the input arguments from the selected "
                                          "test."))
    parser_export.add_argument("test_number", type=int, help=_("number of the test "
                                                               "to export as returned "
                                                               "by the 'info' command."))
    parser_export.add_argument("file", help=_("file containing the tests to be exported."))
    parser_export.set_defaults(config='', timeout=10, verbosity=3, errors=4, color='AUTO', valgrind=False,
                               format='text', log=_DEFAULT_LOG_FILE, test=None, program=None, program_arguments=None,
                               info=False, output_limit=10000)

    return argparser


def check_float_non_negative(value):
    fvalue = float(value)
    if fvalue < 0:
        raise argparse.ArgumentTypeError((_("Invalid parameter"), "('%s')" % value))
    return fvalue


def check_int_non_negative(value):
    ivalue = int(value)
    if ivalue < 0:
        raise argparse.ArgumentTypeError((_("Invalid parameter"), "('%s')" % value))
    return ivalue


def check_int_greater_than_one(value):
    ivalue = int(value)
    if ivalue < 1:
        raise argparse.ArgumentTypeError((_("Invalid parameter"), "('%s')" % value))
    return ivalue


def parse_file(filename):
    """Read the content of a file containing tests."""
    if filename == "":
        return []
    try:
        with open(filename, "rt") as f:
            return list(pvcheck.parser.parse_sections(f))
    except FileNotFoundError as e:
        print(e)
        sys.exit(1)


def test_names_list(test_suite):
    """Build a list containing all the test names of a test suite."""
    test_names = []
    for test in test_suite.test_cases():
        test_names.append(test.description)
    return test_names


def print_test_names_list(test_suite):
    """Print a list containing all the test names of a test suite."""
    test_names = test_names_list(test_suite)
    test_number = 1
    number_of_spaces = int(math.log10(len(test_names)) + 1)
    for test_name in test_names:
        try:
            print('{:{align}{width}}'.format(str(test_number), align='>', width=number_of_spaces) + ' ' + test_name)
        except TypeError:
            print('{:{align}{width}}'.format(str(test_number), align='>', width=number_of_spaces) + ' ' + 'NoName')
        test_number += 1
    exit(0)


def main():
    """Setup the environment and starts the test session."""
    (args, opts) = parse_options()

    single_test_index = None

    cfg = parse_file(opts["config"])

    if opts["list"]:
        td = parse_file(args["test_file"])
        suite = pvcheck.testdata.TestSuite(cfg + td)
        print_test_names_list(suite)
        exit(0)

    td = parse_file(args["test_file"])

    if opts['valgrind']:
        cfg.append(pvcheck.testdata.Section('VALGRIND', []))

    suite = pvcheck.testdata.TestSuite(cfg + td)

    if opts['run'] is not None and opts['export'] is None:
        single_test_index = opts['run'] - 1
        suite = suite.test_case(single_test_index)

    if opts['export'] is not None:
        single_test_index = opts['export'] - 1
        suite = suite.test_case(single_test_index)
        pvcheck.exporter.export(suite, single_test_index)

    execlass = (pvcheck.valgrind.ValgrindExecutor if opts["valgrind"]
                else pvcheck.executor.Executor)
    exe = execlass()

    if opts["format"] == "interactive":
        fmt = pvcheck.interactiveformatter.InteractiveFormatter()
    elif opts["format"] == "json":
        fmt = pvcheck.jsonformatter.JSONFormatter(indent=4, test_file=args["test_file"])
    elif opts["format"] == "csv":
        fmt = pvcheck.csvformatter.CSVFormatter()
    elif opts["format"] == "html":
        fmt = pvcheck.htmlformatter.HTMLFormatter()
    elif opts["format"] == "text":
        fmtclass = (pvcheck.formatter.ColoredTextFormatter if opts["color"]
                    else pvcheck.formatter.TextFormatter)
        fmt = fmtclass(verbosity=opts["verbosity"],
                       maxerrors=opts["maxerrors"])
    else:
        raise ValueError("Unknown format '{}'".format(opts["format"]))
    # Pvcheck returns as exit code the number of failed tests.
    # 255 represents a generic error.
    retcode = 255
    program = [args['program']]
    if args['program_arguments'] is not None:
        program.extend(args['program_arguments'])

    with open(opts["logfile"], "at") as logfile:
        logfmt = pvcheck.jsonformatter.JSONFormatter(logfile,
                                             test_file=args["test_file"])
        combfmt = pvcheck.formatter.CombinedFormatter([fmt, logfmt])
        pvc = pvcheck.pvcheck.PvCheck(exe, combfmt)
        try:
            if single_test_index is None:
                failures = pvc.exec_suite(suite, program,
                                          timeout=opts["timeout"],
                                          output_limit=opts["output_limit"])
            else:
                failures = pvc.exec_single_test(suite, program,
                                                timeout=opts["timeout"],
                                                output_limit=opts["output_limit"])
            retcode = min(failures, 254)
        finally:
            # in case of exception (e.g. tested a non executable file) write a
            # newline to the json log
            logfile.write("\n")
    sys.exit(retcode)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Interactive interface."""

import curses
import threading
import pvcheck.formatter
import pvcheck.executor
import functools
import pvcheck.i18n

_ = pvcheck.i18n.translate


# TODO: i18n
# TODO: refactoring with Formatter + separate UI class


HELP_MESSAGE = """

PVCHECK automatic verification of computer programs
===================================================


Keybord controls:

  RIGHT    switch to the next test in the suite
  
  LEFT     switch to the previous test in the suite

  DOWN, n  scroll one line down

  UP, p    scroll one line up

  PgDN     scroll one page down
  
  PgUP     scroll one page up
  
  HOME     scroll to the beginning
  
  END      scroll to the end
  
  r        redraw the screen
  
  q, ESC   close the program
  
  s        show a summary of the results
  
  i        show info about the test case
  
  o        show the program's output
  
  h, ?     show this page  

"""

HELP_MESSAGE_IT = """

PVCHECK verifica automatica di programmi 
===================================================


Controlli tramite tastiera

  DESTRA   passa al test successivo nella suite

  SINISTRA passa al test precedente nella suite

  GIU`, n  scorre una riga in giu`

  SU, p    scorre una riga in su

  PgDN     scorre una pagina in giu`

  PgUP     scorre una pagina in su

  HOME     scorre all'inizio

  END      scorre alla fine

  r        ridisegna lo schermo

  q, ESC   chiude il programma
  
  s        mostra un riepilogo dei risultati
  
  i        mostra informazioni sul caso di test
  
  o        mostra l'output del programma

  h, ?     mostra questa pagina

"""


pvcheck.i18n.register_translation(HELP_MESSAGE, "it", HELP_MESSAGE_IT)


_CALLBACKS = {}


def _register_key(*keys):
    """Decorator registering key->callback pairs.

    Keys can be strings or character codes."""
    def decorator(f):
        for k in keys:
            if isinstance(k, str):
                k = ord(k)
            _CALLBACKS[k] = f
        return f
    return decorator


def _synchronized(f):
    """Makes the calls to the method acquire a lock owned by the class instance."""
    @functools.wraps(f)
    def decorated(self, *args, **kwargs):
        with self._mutex:
            return f(self, *args, **kwargs)
    return decorated


class InteractiveFormatter(pvcheck.formatter.Formatter):
    """Formatter that uses curses to report the result."""

    FOOTER_H = 2  # height of the footer
    MAX_W = 512   # max line length

    COLOR_OK = 1
    COLOR_WARN = 2
    COLOR_ERR = 3

    _RESULT_TABLE = {
        pvcheck.executor.ER_OK: None,
        pvcheck.executor.ER_TIMEOUT: _("TIMEOUT EXPIRED: PROCESS TERMINATED"),
        pvcheck.executor.ER_OUTPUT_LIMIT: _("TOO MANY OUTPUT LINES"),
        pvcheck.executor.ER_SEGFAULT: _("PROCESS ENDED WITH A FAILURE (SEGMENTATION FAULT)"),
        pvcheck.executor.ER_ERROR: ("PROCESS ENDED WITH A FAILURE (ERROR CODE {status})"),
        pvcheck.executor.ER_NOTFILE: _("FAILED TO RUN THE FILE '{progname}' the file does not exist)")
    }

    def __init__(self):
        """Create the interactive formatter."""
        self._reports = []
        self._report_index = 0
        self._screen = None
        self._mutex = threading.Lock()
        self._initialization_barrier = None
        self._sections = []
        self._err_counts = {}
        self._warn_counts = {}
        self._ok_counts = {}
        self._error_count = 0
        self._warn_count = 0
        self._ok_count = 0

    @_register_key("q", "Q", 27)  # 27 -> ESC
    def _quit(self):
        self._stop = True

    @_register_key("p", "P", curses.KEY_UP)
    def scroll_line_up(self):
        self._reports[self._report_index].scroll(self._text_height(), -1)
        self._update()

    @_register_key("n", "N", curses.KEY_DOWN, 10)  # 10 -> ENTER
    def scroll_line_down(self):
        self._reports[self._report_index].scroll(self._text_height(), 1)
        self._update()

    @_register_key(curses.KEY_PPAGE)
    def scroll_page_up(self):
        self._reports[self._report_index].scroll(self._text_height(), pages=-1)
        self._update()

    @_register_key(curses.KEY_NPAGE)
    def scroll_page_down(self):
        self._reports[self._report_index].scroll(self._text_height(), pages=1)
        self._update()

    @_register_key(curses.KEY_HOME)
    def scroll_begin(self):
        self._reports[self._report_index].scroll(self._text_height(), documents=-1)
        self._update()

    @_register_key(curses.KEY_END)
    def scroll_end(self):
        self._reports[self._report_index].scroll(self._text_height(), documents=1)
        self._update()

    @_register_key("r", "R", curses.KEY_RESIZE)
    def resize_terminal(self):
        self._update()

    @_register_key(curses.KEY_LEFT)
    def previous_report(self):
        self._screen.clear()
        if len(self._reports) > 1:
            self._report_index = max(self._report_index - 1, 1)
        self._update()

    @_register_key(curses.KEY_RIGHT)
    def next_report(self):
        self._screen.clear()
        self._report_index = min(self._report_index + 1, len(self._reports) - 1)
        self._update()

    @_register_key("h", "H", "?")
    def next_report(self):
        self._screen.clear()
        self._show_info(_(HELP_MESSAGE))
        self._screen.clear()
        self._update()

    @_register_key("i", "I")
    def next_report(self):
        self._screen.clear()
        doc = self._reports[self._report_index]
        self._show_info(doc.info())
        self._screen.clear()
        self._update()

    @_register_key("o", "O")
    def next_report(self):
        self._screen.clear()
        doc = self._reports[self._report_index]
        text = _("PROGRAM'S OUTPUT:")
        self._show_info(text + "\n" + doc.output)
        self._screen.clear()
        self._update()

    @_register_key("s", "S")
    def next_report(self):
        self._screen.clear()
        doc = self._reports[self._report_index]
        text = [_("SUMMARY:"), ""]
        for s in self._sections:
            line = "%20s %3d ok  %3d warnings  %3d errors" % (s, self._ok_counts[s], self._warn_counts[s], self._err_counts[s])
            text.append(line)
        text.append(" ")
        self._show_info("\n".join(text))
        self._screen.clear()
        self._update()

    def _thread_body(self):
        """UI thread."""
        curses.wrapper(self._main_loop)

    def _main_loop(self, screen):
        """Main loop managing the interaction with the user."""
        # First setup curses
        self._screen = screen
        curses.use_default_colors()
        curses.init_pair(self.COLOR_OK, curses.COLOR_GREEN, -1)
        curses.init_pair(self.COLOR_WARN, curses.COLOR_YELLOW, -1)
        curses.init_pair(self.COLOR_ERR, curses.COLOR_RED, -1)
        self._footer = curses.newwin(self.FOOTER_H, self._text_width(), self._text_height(), 0)
        self._footer.bkgd(" ", curses.A_REVERSE)
        self._reports.append(Report("PVCHECK", self.MAX_W))
        self._reports[-1].add_line("Waiting for test results...")
        self._report_index = 0
        self._update()
        self._stop = False
        # This reactivate the main thread
        self._initialization_barrier.wait()
        # Event loop
        while not self._stop:
            ch = screen.getch()
            with self._mutex:
                _CALLBACKS.get(ch, lambda self: None)(self)

    def _text_height(self):
        """Number of text lines displayed."""
        return self._screen.getmaxyx()[0] - self.FOOTER_H

    def _text_width(self):
        """Width of the displayed text."""
        return min(self._screen.getmaxyx()[1], self.MAX_W)

    def _add_footer(self, line, align, text, *extra):
        """Add some text in the footer."""
        k = self._text_width() - 1 - len(text)
        pos = max(0, (0 if align == "left" else (k if align == "right" else k //2 )))
        self._footer.addnstr(line, pos, text, self._text_width() - 1 - pos, *extra)

    def _add_short_report(self):
        """Insert ok, warnings, errors counters in the footer."""
        texts = [
            "%3d " % self._ok_count, " %s, " % _("passes"),
            "%3d " % self._warn_count, " %s, " % _("warnings"),
            "%3d " % self._err_count, " %s" % _("errors")
        ]
        styles = [
            [curses.color_pair(self.COLOR_OK)], [],
            [curses.color_pair(self.COLOR_WARN)], [],
            [curses.color_pair(self.COLOR_ERR)], []
        ]
        n = self._text_width() - 1
        pos = 0
        for t, s in zip (texts, styles):
            self._footer.addnstr(1, pos, t, n, *s)
            pos += len(t)
            n -= len(t)
        
    def _update(self):
        """Redraw everything."""
        if self._screen is None:
            return
        self._screen.refresh()
        height = self._text_height()
        width = self._text_width()
        doc = self._reports[self._report_index]
        self._footer.mvwin(height, 0)
        doc.refresh(self._text_height(), self._text_width())
        self._footer.clear()
        self._add_footer(0, "center", _("[Press 'h' for help]"), curses.A_DIM)
        text = _("Test case %d of %d (%s)") % (self._report_index, len(self._reports) - 1, doc.title) + " "
        self._add_footer(0, "left", text)
        text = _("Lines %d-%d/%d") % (doc.top(), doc.bottom(height), doc.length())
        self._add_footer(0, "right", text)
        self._add_short_report()
        if self._running:
            tot = self._err_count + self._warn_count + self._ok_count
            text = _("TEST RUNNING") + "..." + "|/-\\"[tot % 4]
        else:
            text = _("TEST COMPLETED")
        self._add_footer(1, "right", text, curses.A_BOLD)
        self._footer.refresh()

    def _show_info(self, text):
        """Show some text on the screen temporarily disabling the main interface."""
        self._screen.refresh()
        lines = text.splitlines()
        content_pad = curses.newpad(len(lines), 1 + max(map(len, lines)))
        for n, line in enumerate(lines):
            content_pad.addstr(n, 0, line)
        start_line = 0
        while True:
            height, width = self._screen.getmaxyx()
            start_line = max(0, start_line)
            start_line = min(len(lines) - height, start_line)
            content_pad.refresh(start_line, 0, 0, 0, height - 1, width - 1)
            ch = self._screen.getch()
            if ch in (curses.KEY_DOWN, ord("n"), ord("N")):
                start_line += 1
            elif ch in (curses.KEY_UP, ord("p"), ord("P")):
                start_line -= 1
            elif ch == curses.KEY_NPAGE:
                start_line += height
            elif ch == curses.KEY_PPAGE:
                start_line -= height
            elif ch == curses.KEY_END:
                start_line += len(lines)
            elif ch == curses.KEY_HOME:
                start_line = 0
            else:
                break

    # -- Formatter interface --------------------------------------------------
        
    def begin_session(self):
        self._err_count = self._warn_count = self._ok_count = 0
        self._running = True
        # Start the UI thread
        self._initialization_barrier = threading.Barrier(2)
        self._thread = threading.Thread(target=self._thread_body)
        self._thread.start()
        self._initialization_barrier.wait()

    def end_session(self):
        # Wait the termination of the UI thread
        with self._mutex:
            self._running = False
            self._update()
        self._thread.join()

    @_synchronized
    def begin_test(self, description, cmdline_args, input, tempfile):
        description = description or ""
        self._reports.append(Report(description, self.MAX_W, cmdline_args, input, tempfile))
        if self._report_index == 0:
            self._report_index = 1
        self._update()

    @_synchronized
    def end_test(self):
        pass

    @_synchronized
    def execution_result(self, cmdline_args, execution_result, test):
        info = {
            'progname': cmdline_args[0],
            'status': execution_result.status
        }
        message = self._RESULT_TABLE[execution_result.result]
        if execution_result.result != pvcheck.executor.ER_OK:
            self._err_count += 1
            message = message.format(**info)
            for line in message.splitlines():
                self._reports[-1].add_line(line, curses.color_pair(self.COLOR_ERR))
        self._reports[-1].output = execution_result.output
        self._update()

    def _new_section(self, section_name):
        if section_name not in self._sections:
            self._sections.append(section_name)
            self._err_counts[section_name] = 0
            self._warn_counts[section_name] = 0
            self._ok_counts[section_name] = 0

    @_synchronized
    def comparison_result(self, expected, got, diffs, matches):
        add = self._reports[-1].add_line
        all_ok = (max(diffs, default=0) <= 0)
        self._new_section(expected.tag)
        if all_ok:
            self._ok_count += 1
            self._ok_counts[expected.tag] += 1
        else:
            self._err_count += 1
            self._err_counts[expected.tag] += 1
        color = curses.color_pair(self.COLOR_OK if all_ok else self.COLOR_ERR)
        add("[%s]" % expected.tag, color | curses.A_BOLD)

        err_diff = "%s\t\t(" +_("expected") + " '%s')"
        err_un = "%s\t\t(" + _("this line was not expected") + ")"
        err_mis = "\t\t(" + _("missing line") + " '%s'"

        for (i, d) in enumerate(diffs):
            try:
                out_string = pvcheck.formatter.handle_non_printable_chars(got.content[i])
            except IndexError:
                out_string = ''
            if d <= 0:
                # Correct
                add(out_string, curses.color_pair(self.COLOR_OK))
            elif matches[i] is None:
                # Extra line
                add(err_un % out_string, curses.color_pair(self.COLOR_ERR))
            elif i >= len(got.content):
                # Missing line
                add(err_mis % matches[i], curses.color_pair(self.COLOR_ERR))
            else:
                # Mismatching lines

                add(err_diff % (out_string, matches[i]), curses.color_pair(self.COLOR_ERR))
        add("")
        self._update()

    @_synchronized
    def missing_section(self, expected):
        self._new_section(expected.tag)
        self._warn_count += 1
        self._warn_counts[expected.tag] += 1
        message = ("\t\t(" + _("section [%s] is missing") + ")") % expected.tag
        self._reports[-1].add_line(message, curses.color_pair(self.COLOR_WARN))
        self._reports[-1].add_line("")
        self._update()


class Report:
    """Test result displayed on the screen."""

    def __init__(self, title, max_width, cmdline_args="", input="", tempfile=""):
        """Create a new report with the given title.

        Lines are truncated to max_width.
        """
        self.title = title
        self._length = 0
        self._max_width = max_width
        self._pad = curses.newpad(1, max_width)
        self._position = 0
        self._cmdline_args = cmdline_args
        self._input = input
        self._tempfile = tempfile
        self.output = ""

    def add_line(self, line, *extra):
        """Add a line at the bottom of the document."""
        self._pad.resize(self._length + 1, self._max_width)
        self._pad.addnstr(self._length, 0, line, self._max_width, *extra)
        self._length += 1

    def scroll(self, page_height, lines=0, pages=0, documents=0):
        """Scroll up or down."""
        amount = lines + page_height * pages + documents * self._length
        self._position = max(0, min(self._position + amount, self._length - page_height))

    def refresh(self, page_height, page_width):
        """Redraw the document on the screen."""
        self._pad.refresh(self._position, 0, 0, 0, page_height - 1, page_width - 1)

    def top(self):
        """Index of the first line on the screen (starting from 1)."""
        return self._position + 1

    def bottom(self, page_height):
        """Index of the last line on the screen."""
        return min(self._position + page_height, self._length)

    def length(self):
        """Number of lines in the document."""
        return self._length

    def info(self):
        lines = [_("Test title: %s") % self.title]
        # , "", _("Command line: %s") % " ".join(self._cmdline_args)]
        if self._cmdline_args:
            lines.append("")
            args = [_("TEMP_FILE") if x is pvcheck.executor.ARG_TMPFILE else x for x in self._cmdline_args]
            lines.append(_("Command line: %s") % " ".join(args))
        if self._input and self._input.strip():
            lines.extend(["", _("Input:")])
            lines.extend(self._input.splitlines())
        if self._tempfile:
            lines.extend(["", _("Temporary file:")])
            lines.extend(self._tempfile.splitlines())
        return "\n".join(lines)



import tempfile
import logging
import subprocess

_log = logging.getLogger()


def parse_wdiff_output(wd_str):
    """Generic function to parse wdiff -s output"""
    tokens = wd_str.split()
    n1 = float(tokens[4][:-1])
    n2 = float(tokens[15][:-1])
    return (n1 + n2) / 2


def check_wdiff(res, target):
    file_text = ("\n".join(res))
    tmpfile1 = tempfile.NamedTemporaryFile(suffix=".res.pvcheck.tmp",
                                          delete=False)
    tmpfile1.write(file_text.encode('utf8'))
    tmpfile1.close()

    file_text = ("\n".join(target))
    tmpfile2 = tempfile.NamedTemporaryFile(suffix=".target.pvcheck.tmp",
                                          delete=False)
    tmpfile2.write(file_text.encode('utf8'))
    tmpfile2.close()

    ret = {}
    try:
        p = subprocess.Popen(['wdiff', '-s123', tmpfile1.name,
                              tmpfile2.name],
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, bufsize=-1)
        output, error = p.communicate()
        ret['code'] = p.returncode
        if p.returncode <= 1:
           ret['output'] = str(output)
        else:
           ret['output'] = "Error occurred in wdiff"
        _log.debug("wdiff: " + str(((output, p.returncode))))
    except FileNotFoundError:
        _log.debug("wdiff: command not found")
        ret['code'] = -1
        ret['output'] = "wdiff command not found"

    return ret



"""Formatter producing JSON data."""
import os
import sys
import json
import datetime
from collections import OrderedDict
from itertools import zip_longest
import pvcheck.formatter
import pvcheck.executor


JSON_FORMAT_VER = "2.2.0"


# TO BE DEFINED
# - wrong lines
# - return_code/code/section status
# - difference/similarity


class JSONFormatter(pvcheck.formatter.Formatter):
    _RESULT_TABLE = {
        pvcheck.executor.ER_OK: "ok",
        pvcheck.executor.ER_TIMEOUT: "TIMEOUT EXPIRED: PROCESS TERMINATED",
        pvcheck.executor.ER_OUTPUT_LIMIT: "TOO MANY OUTPUT LINES",
        pvcheck.executor.ER_SEGFAULT:
        "PROCESS ENDED WITH A FAILURE (SEGMENTATION FAULT)",
        pvcheck.executor.ER_ERROR:
        "PROCESS ENDED WITH A FAILURE (ERROR CODE {status})",
        pvcheck.executor.ER_NOTFILE:
        "FAILED TO RUN THE FILE '{progname}' (the file does not exist)"
    }

    def __init__(self, destination=sys.stdout, indent=None, test_file=None):
        self._dest = destination
        self._indent = indent
        self._obj = None
        self._tests = []
        self._work_dir = os.getcwd()
        self._test_file = self._work_dir
        if test_file is not None:
            self._test_file += "/" + test_file
        else:
            self._test_file += "/pvcheck.test"

    def _now(self):
        now = datetime.datetime.now()
        (dt, micro) = now.strftime("%Y-%m-%d %H:%M:%S.%f").split(".")
        return "%s.%03d" % (dt, int(micro) / 1000)

    def _proc_args(self, args):
        return [(a if a is not pvcheck.executor.ARG_TMPFILE
                 else "<temp.file>")
                for a in args]
    
    def begin_session(self):
        self._tests = []
        self._obj = OrderedDict([
            ("created_at", self._now()),
            ("version", JSON_FORMAT_VER),
            ("working_directory", self._work_dir),
            ("test_file", self._test_file),
            ("tests", self._tests)
        ])

    def end_session(self):
        json.dump(self._obj, self._dest, indent=self._indent)

    def begin_test(self, description, cmdline_args, input, tempfile):
        t = OrderedDict([
            ("title", description),
            ("command_line", self._proc_args(cmdline_args)),
            ("input_text", input),
            ("file_text", tempfile),
            ("input_file_name",      # not available here
             (None if tempfile is None else "<temp.file>"))
        ])
        self._tests.append(t)

    def execution_result(self, cmdline_args, execution_result, test):
        t = self._tests[-1]
        info = {
            "progname": cmdline_args[0],
            "status": execution_result.status
        }
        msg = self._RESULT_TABLE[execution_result.result]
        t["return_code"] = execution_result.status
        t["error_message"] = msg.format(**info)
        t["output"] = execution_result.output
        self._sections = OrderedDict()
        if execution_result.result != pvcheck.executor.ER_OK:
            for s in test.sections(exclude_special=True):
                self._sections[s.tag] = OrderedDict([("section status",
                                                      "exec_error")])

        t["sections"] = self._sections

    def comparison_result(self, expected, got, diffs, matches):
        status = ("ok" if max(diffs, default=0) == 0
                  else "error")   # ???

        wrong = [(n, a, b) for (n, a, b, d)
                 in zip_longest(range(len(diffs)), got.content,
                                matches, diffs)
                 if d > 0]
            
        s = OrderedDict([
            ("section status", status),
            ("expected", expected.content),
            ("generated", got.content),
            ("wrong_lines", wrong),
	    ("difference", sum(diffs))
        ])
        self._sections[expected.tag] = s
        
    def missing_section(self, expected):
        s = OrderedDict([("section status", "missing")])
        self._sections[expected.tag] = s



"""Main PvCheck class."""


import pvcheck.match
import pvcheck.parser
import pvcheck.executor


class PvCheck:
    """Main class that runs the tests."""

    def __init__(self, executor, formatter):
        self._exec = executor
        self._fmt = formatter

    def exec_suite(self, suite, args, timeout=None, output_limit=None):
        """Verify the program with a collection of test cases.

        Return the number of failed tests.
        """
        self._fmt.begin_session()
        failures = 0
        try:
            for test in suite.test_cases():
                if not self._exec_test(test, args, timeout=timeout,
                                       output_limit=output_limit):
                    failures += 1
                self._fmt.end_test()
        finally:
            self._fmt.end_session()
        return failures

    def exec_single_test(self, test, args, timeout=None, output_limit=None):
        """Verify the program on a single test case.

        Return True if the test has been successfully passed.
        """
        self._fmt.begin_session()
        success = False
        try:
            success = self._exec_test(test, args, timeout=timeout,
                                      output_limit=output_limit)
        finally:
            self._fmt.end_test()
            self._fmt.end_session()
        return success

    def _exec_test(self, test, args, timeout=None, output_limit=None):
        # Run the program and verify it according to the test case.
        # Return True if the test is successful.
        input = test.find_section_content(".INPUT", "")
        tmpfile = test.find_section_content(".FILE", None)
        args = list(args)
        arg_sect = test.find_section(".ARGS")
        if arg_sect is not None:
            args.extend(map(str.strip, arg_sect.content))
            if tmpfile is not None:
                args = [(a if a != ".FILE" else pvcheck.executor.ARG_TMPFILE)
                        for a in args]

        self._fmt.begin_test(test.description, args, input, tmpfile)

        exec_result = self._exec.exec_process(
            args, input, tmpfile=tmpfile,
            timeout=timeout,
            output_limit=output_limit
        )
        self._fmt.execution_result(args, exec_result, test)
        if exec_result.result == pvcheck.executor.ER_OK:
            return self._check_output(test, exec_result.output)
        else:
            return False

    def _check_output(self, test, output):
        # Return True if the test has been passed.
        success = True
        answers = list(pvcheck.parser.parse_sections(output.splitlines()))
        for s in test.sections(exclude_special=True):
            for ans in answers:
                if ans.tag != s.tag:
                    continue
                ordered = ("unordered" not in test.section_options(s.tag))
                diffs, matches = pvcheck.match.compare_sections(
                    ans.content, s.content, ordered=ordered
                )
                self._fmt.comparison_result(s, ans, diffs,
                                            matches)
                success = success and (max(diffs, default=0.0) == 0)
                break
            else:
                success = False
                self._fmt.missing_section(s)
        return success



"""Parse text representing sections."""

import re
import pvcheck.testdata


# Any character is allowed before '[', otherwise the header will be
# missed after a scanf (or other input) where the newline is taken
# from the stdin and is not part of the stdout.  In practice, the
# condition to deal with is 
#
# ...
# printf("Enter n: ");
# scanf("%d", &n);
# printf("[SECTION]\n%d\n", n);
# ...
#
# whose output is:
#
# Enter n:  [SECTION]
# 42
#
_RE_HEADER = re.compile(r"[^[]*\[\s*(?P<tag>[-._a-zA-Z][-._\w]*)\s*\]\s*")


def parse_sections(f):
    """Parse the lines in f and generates sections.

    When non-empty lines are found before the first section header,
    they are placed into a section with an empty name.

    """
    content = []
    tag = ""
    for line in f:
        l = line.strip()
        if not l or l[0] == '#':
            continue  # Skip empty lines and comments
        m = _RE_HEADER.match(line)
        if m:
            if tag != "" or len(content) > 0:
                yield pvcheck.testdata.Section(tag, content)
            # A new section
            tag = m.group("tag")
            content = []
        else:
            content.append(line.rstrip())
    if tag != "" or len(content) > 0:
        yield pvcheck.testdata.Section(tag, content)



from pvcheck.i18n import translate as _


def export(test, test_index):
    """Export input data from a test."""
    output_file_name = _file_name_builder(test.description)
    _input, _file = _get_data(test)
    if (_input is not None) or (_file is not None):
        with open(output_file_name, 'w') as f:
            _write_data(f, _input, _file)
        exit(0)
    else:
        _print_error_message(test_index)
        exit(1)


def _file_name_builder(test_name):
    """Build a name for the output file."""
    try:
        file_name = (test_name + ".dat").replace(' ', '_')
    except TypeError:
        file_name = ("NoName" + ".dat")
    return file_name


def _get_data(test):
    """Obtain data from certain sections of a test.
    
    Sections: - [.INPUT]
              - [.FILE]
    
    """
    _input = test.find_section_content(".INPUT", None)
    _file = test.find_section_content(".FILE", None)
    return _input, _file


def _write_data(output_file, _input, _file):
    """Write data into the output file."""
    if _input is not None:
        output_file.write(_input)
    if _file is not None:
        output_file.write(_file)


def _print_error_message(test_index):
    """Print an error message containing the test number."""
    fmt = _("Error: Can't export test number %d.")
    msg = fmt % (test_index + 1)
    print("\n" + msg + "\n")



"""Executor using the valgrind memory checker tool."""

import pvcheck.executor


class ValgrindExecutor(pvcheck.executor.Executor):
    """Executor that uses valgrind to check the memory usage.
    
    Output from valgrind is detected and used to form a new 'VALGRIND'
    section at the end of the regular process output.  In absence of
    errors that section would be empty.

    Valgrind need to be installed in the system.

    """
    
    def exec_process(self, args, *rest, **kwargs):
        res = super().exec_process(["valgrind"]+args, *rest, **kwargs)
        out, err = self._process_output(res.output, res.stderr)
        return res._replace(output=out, stderr=err)

    def _process_output(self, stdout, stderr):
        extraout = ["\n[VALGRIND]\n"]
        newerr = []
        for line in stderr.splitlines(keepends=True):
            if not line.startswith("=="):
                newerr.append(line)
                continue
            if "in use at exit" in line:
                f = line.split()
                bytes_ = int(f[5].replace(',', ''))
                if bytes_ > 0:
                    extraout.append(line)
            elif "total heap usage" in line:
                f = line.split()
                allocs = int(f[4].replace(',', ''))
                frees = int(f[6].replace(',', ''))
                if allocs != frees:
                    extraout.append(line)
            elif "ERROR SUMMARY" in line:
                f = line.split()
                errs = int(f[3].replace(',', ''))
                if errs > 0:
                    extraout.append(line)

        newout = stdout + "".join(extraout)
        return (newout, "".join(newerr))



"""Execution of the child process."""


import subprocess
import contextlib
import tempfile
import signal
import os
import collections


# Execution results
ER_OK = "ER_OK"
ER_TIMEOUT = "ER_TIMEOUT"
ER_OUTPUT_LIMIT = "ER_OUTPUT_LIMIT"
ER_SEGFAULT = "ER_SEGFAULT"
ER_ERROR = "ER_ERROR"
ER_NOTFILE = "ER_NOTFILE"

# Placeholder for the name of the temporary file
ARG_TMPFILE = object()


@contextlib.contextmanager
def _make_temp_file(content):
    tmp = tempfile.NamedTemporaryFile(mode="wt",
                                      suffix=".pvcheck.tmp",
                                      delete=False)
    name = tmp.name
    try:
        tmp.write(content)
        tmp.close()
        yield name
    finally:
        os.remove(name)


ExecResult = collections.namedtuple(
    'ExecResult', ['result', 'status', 'output', 'stderr']
)


class Executor:
    """Class capable of executing a process."""
    
    def exec_process(self, args, input, tmpfile=None, timeout=None,
                     output_limit=None):
        """Execute a process.

        Parameters:
        args    - list of arguments (the first being the process name)
        input   - text to be sent on the standard input
        tmpfile - optional temporary file (can be None)
        timeout - in seconds, None for unlimited time
        output_limit - maximum number of output lines, None if unlimited

        When one of the arguments is the placehoder 'ARG_TMPFILE' it
        gets replaced by the name of a temporary file having the
        content specified by the tmpfile parameter.

        Return an ExecResult object, that is, a named tuple with the
        result of the execution, the status code of the terminated
        process, and the output produced by the process.

        The output limit is applied independently to stdout and stderr.
        When exceeded the execution is considered as failed.

        """
        with contextlib.ExitStack() as stack:
            if tmpfile is not None:
                tmpname = stack.enter_context(_make_temp_file(tmpfile))
                args = self._replace_placeholder(args, tmpname)
            outputb = b""
            errorb = b""
            ret_code = 0
            try:
                proc = subprocess.Popen(args,
                                        stdin=subprocess.PIPE,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE)
                inputb = input.encode('utf-8', errors='ignore')
                outputb, errorb = proc.communicate(inputb, timeout)
                codes = {0: ER_OK, -signal.SIGSEGV: ER_SEGFAULT}
                er = codes.get(proc.returncode, ER_ERROR)
                ret_code = proc.returncode
            except subprocess.TimeoutExpired:
                er = ER_TIMEOUT
                proc.kill()
                proc.stdin.close()
                proc.stdout.close()
                proc.stderr.close()
                proc.wait()
            except FileNotFoundError:
                er = ER_NOTFILE
        output = outputb.decode('utf-8', errors='ignore')
        error = errorb.decode('utf-8', errors='ignore')
        if output_limit is not None:
            lines = output.splitlines(True) 
            if len(lines) > output_limit:
                output = "".join(lines[:output_limit])
                er = ER_OUTPUT_LIMIT
            lines = error.splitlines(True) 
            if len(lines)  > output_limit:
                error = "".join(lines[:output_limit])
                er = ER_OUTPUT_LIMIT
        return ExecResult(er, ret_code, output, error)

    def _replace_placeholder(self, args, name):
        return [(name if a is ARG_TMPFILE else a) for a in args]





"""Classes defining test cases and test suites."""

from collections import OrderedDict
from itertools import chain
from pvcheck.i18n import translate as _


class Section:
    """A section denoted by a tag and with some content."""
    def __init__(self, tag, content):
        self.tag = tag
        self.content = content

    def copy(self):
        return Section(self.tag, self.content[:])

    def text(self):
        if not self.content:
            return ""
        else:
            return "\n".join(self.content) + "\n"

    def __repr__(self):
        return ("Section('%s')" % self.tag)


class TestCase:
    """A test case with an optional description."""
    def __init__(self, description=None, sections=()):
        """Create a new test case with the given description."""
        self.description = description
        self._sections = OrderedDict()
        self._section_options = {}
        for s in sections:
            self.add_section(s)

    def add_section(self, section):
        """Add a section to the test case.

        If a section with the same tag already exists, then the
        content of the two are merged.  The content of the special
        '.SECTIONS' section is parsed and recorded as the options of
        the test case.

        """
        if section.tag == ".SECTIONS":
            opts = self._parse_section_options(section.content)
            self._section_options.update(opts)
        elif section.tag in self._sections:
            self._sections[section.tag].content.extend(section.content)
        else:
            self._sections[section.tag] = section.copy()

    def find_section(self, tag):
        """Search for a section with the given tag name."""
        return self._sections.get(tag)

    def find_section_content(self, tag, default=""):
        """Search for a section with the given tag name."""
        s = self._sections.get(tag)
        return (s.text() if s is not None else default)

    def section_options(self, tag):
        """Return the special options for a section."""
        return self._section_options.get(tag, set())

    def sections(self, exclude_special=False):
        """Return the sections in the test case.

        Special sections can be excluded.

        """
        if exclude_special:
            return [s for s in self._sections.values() if s.tag and not
                    s.tag.startswith(".")]
        else:
            return list(self._sections.values())

    def _parse_section_options(self, lines):
        opts = dict((s[0], set(s[1:]))
                    for s in (l.split() for l in lines) if s)
        return opts


class TestSuite:
    """A set of test cases.

    Includes an optional prefix of sections, that are replicated for
    each case.

    """

    def __init__(self, sections=()):
        """Create a suite from a sequence of sections."""
        it = self._group_tests(sections)
        prefix = next(it, (None, []))[1]

        self._cases = []
        for testname, secs in it:
            c = TestCase(testname)
            for s in chain(prefix, secs):
                c.add_section(s)
            self._cases.append(c)
        if not self._cases:
            # When no test is found, the prefix becomes a single
            # unnamed test case.
            c = TestCase()
            for s in prefix:
                c.add_section(s)
            self._cases.append(c)

    def test_cases(self):
        """Return the list of test cases in the suite."""
        return self._cases

    def test_case(self, test_number):
        """Return a test case."""
        try:
            if test_number < 0:
                raise IndexError
            test_case = self.test_cases()[test_number]
        except IndexError:
            fmt = _("Test number %d doesn't exist.")
            msg = fmt % (test_number + 1)
            print("\n" + msg + "\n")
            print(_("Use './pvcheck info' to list all the available tests.") + "\n")
            exit(2)
        return test_case

    def _group_tests(self, sections):
        name = None
        lst = []
        n = 0
        for s in sections:
            if s.tag == '.TEST':
                yield (name, lst)
                lst = []
                n += 1
                name = (s.content[0].strip() if len(s.content) > 0
                        else "Test-%d" % n)
            else:
                lst.append(s)
        if n > 0 or lst:
            yield (name, lst)



"""ArgumentParser with Italian translation."""

import argparse
import pvcheck.i18n
import sys
_ = pvcheck.i18n.translate


def _callable(obj):
    return hasattr(obj, '__call__') or hasattr(obj, '__bases__')


class ArgParser(argparse.ArgumentParser):
    def __init__(self,
                 prog=None,
                 usage=None,
                 description=None,
                 epilog=None,
                 version=None,
                 parents=[],
                 formatter_class=argparse.HelpFormatter,
                 prefix_chars='-',
                 fromfile_prefix_chars=None,
                 argument_default=None,
                 conflict_handler='error',
                 add_help=True):
        object.__init__(self)
        self.set_default_subparser
        self.exit_on_error = True
        if version is not None:
            import warnings
            warnings.warn(
                """The "version" argument to ArgumentParser is deprecated. """
                """Please use """
                """"add_argument(..., action='version', version="N", ...)" """
                """instead""", DeprecationWarning)

        superinit = super(argparse.ArgumentParser, self).__init__
        superinit(description=description,
                  prefix_chars=prefix_chars,
                  argument_default=argument_default,
                  conflict_handler=conflict_handler)

        # default setting for prog
        if prog is None:
            prog = argparse._os.path.basename(argparse._sys.argv[0])

        self.prog = prog
        self.usage = usage
        self.epilog = epilog
        self.version = version
        self.formatter_class = formatter_class
        self.fromfile_prefix_chars = fromfile_prefix_chars
        self.add_help = add_help

        add_group = self.add_argument_group
        self._positionals = add_group(_('positional arguments'))
        self._optionals = add_group(_('optional arguments'))
        self._subparsers = None

        # register types
        def identity(string):
            return string
        self.register('type', None, identity)

        # add help and version arguments if necessary
        # (using explicit default to override global argument_default)
        if self.add_help:
            self.add_argument(
                '-h', '--help', action='help', default=argparse.SUPPRESS,
                help=_('show this help message and exit'))
        if self.version:
            self.add_argument(
                '-v', '--version', action='version', default=argparse.SUPPRESS,
                version=self.version,
                help=_("show program's version number and exit"))

        # add parent arguments and defaults
        for parent in parents:
            self._add_container_actions(parent)
            try:
                defaults = parent._defaults
            except AttributeError:
                pass
            else:
                self._defaults.update(defaults)

    def set_default_subparser(self, name, args=None):
        """
        Default subparser selection.

        name: is the name of the subparser to call by default
        args: if set is the argument list handed to parse_args()
        """
        subparser_found = False
        for arg in sys.argv[1:]:
            if arg in ['-h', '--help']:  # global help if no subparser
                break
        else:
            for x in self._subparsers._actions:
                if not isinstance(x, argparse._SubParsersAction):
                    continue
                for sp_name in x._name_parser_map.keys():
                    if sp_name in sys.argv[1:]:
                        subparser_found = True
            if not subparser_found:
                # insert default in first position, this implies no
                # global options without a sub_parsers specified
                if args is None:
                    sys.argv.insert(1, name)
                else:
                    args.insert(0, name)

    # =====================================
    # Command line argument parsing methods
    # =====================================
    def parse_args(self, args=None, namespace=None):
        args, argv = self.parse_known_args(args, namespace)
        if argv:
            msg = _('unrecognized arguments: %s')
            self.error(msg % ' '.join(argv))
        return args


    def _parse_known_args(self, arg_strings, namespace):
        # replace arg strings that are file references
        if self.fromfile_prefix_chars is not None:
            arg_strings = self._read_args_from_files(arg_strings)

        # map all mutually exclusive arguments to the other arguments
        # they can't occur with
        action_conflicts = {}
        for mutex_group in self._mutually_exclusive_groups:
            group_actions = mutex_group._group_actions
            for i, mutex_action in enumerate(mutex_group._group_actions):
                conflicts = action_conflicts.setdefault(mutex_action, [])
                conflicts.extend(group_actions[:i])
                conflicts.extend(group_actions[i + 1:])

        # find all option indices, and determine the arg_string_pattern
        # which has an 'O' if there is an option at an index,
        # an 'A' if there is an argument, or a '-' if there is a '--'
        option_string_indices = {}
        arg_string_pattern_parts = []
        arg_strings_iter = iter(arg_strings)
        for i, arg_string in enumerate(arg_strings_iter):

            # all args after -- are non-options
            if arg_string == '--':
                arg_string_pattern_parts.append('-')
                for arg_string in arg_strings_iter:
                    arg_string_pattern_parts.append('A')

            # otherwise, add the arg to the arg strings
            # and note the index if it was an option
            else:
                option_tuple = self._parse_optional(arg_string)
                if option_tuple is None:
                    pattern = 'A'
                else:
                    option_string_indices[i] = option_tuple
                    pattern = 'O'
                arg_string_pattern_parts.append(pattern)

        # join the pieces together to form the pattern
        arg_strings_pattern = ''.join(arg_string_pattern_parts)

        # converts arg strings to the appropriate and then takes the action
        seen_actions = set()
        seen_non_default_actions = set()

        def take_action(action, argument_strings, option_string=None):
            seen_actions.add(action)
            argument_values = self._get_values(action, argument_strings)

            # error if this argument is not allowed with other previously
            # seen arguments, assuming that actions that use the default
            # value don't really count as "present"
            if argument_values is not action.default:
                seen_non_default_actions.add(action)
                for conflict_action in action_conflicts.get(action, []):
                    if conflict_action in seen_non_default_actions:
                        msg = _('not allowed with argument %s')
                        action_name = argparse._get_action_name(conflict_action)
                        raise argparse.ArgumentError(action, msg % action_name)

            # take the action if we didn't receive a SUPPRESS value
            # (e.g. from a default)
            if argument_values is not argparse.SUPPRESS:
                action(self, namespace, argument_values, option_string)

        # function to convert arg_strings into an optional action
        def consume_optional(start_index):

            # get the optional identified at this index
            option_tuple = option_string_indices[start_index]
            action, option_string, explicit_arg = option_tuple

            # identify additional optionals in the same arg string
            # (e.g. -xyz is the same as -x -y -z if no args are required)
            match_argument = self._match_argument
            action_tuples = []
            while True:

                # if we found no optional action, skip it
                if action is None:
                    extras.append(arg_strings[start_index])
                    return start_index + 1

                # if there is an explicit argument, try to match the
                # optional's string arguments to only this
                if explicit_arg is not None:
                    arg_count = match_argument(action, 'A')

                    # if the action is a single-dash option and takes no
                    # arguments, try to parse more single-dash options out
                    # of the tail of the option string
                    chars = self.prefix_chars
                    if arg_count == 0 and option_string[1] not in chars:
                        action_tuples.append((action, [], option_string))
                        for char in self.prefix_chars:
                            option_string = char + explicit_arg[0]
                            explicit_arg = explicit_arg[1:] or None
                            optionals_map = self._option_string_actions
                            if option_string in optionals_map:
                                action = optionals_map[option_string]
                                break
                        else:
                            msg = _('ignored explicit argument %r')
                            raise argparse.ArgumentError(action, msg % explicit_arg)

                    # if the action expect exactly one argument, we've
                    # successfully matched the option; exit the loop
                    elif arg_count == 1:
                        stop = start_index + 1
                        args = [explicit_arg]
                        action_tuples.append((action, args, option_string))
                        break

                    # error if a double-dash option did not use the
                    # explicit argument
                    else:
                        msg = _('ignored explicit argument %r')
                        raise argparse.ArgumentError(action, msg % explicit_arg)

                # if there is no explicit argument, try to match the
                # optional's string arguments with the following strings
                # if successful, exit the loop
                else:
                    start = start_index + 1
                    selected_patterns = arg_strings_pattern[start:]
                    arg_count = match_argument(action, selected_patterns)
                    stop = start + arg_count
                    args = arg_strings[start:stop]
                    action_tuples.append((action, args, option_string))
                    break

            # add the Optional to the list and return the index at which
            # the Optional's string args stopped
            assert action_tuples
            for action, args, option_string in action_tuples:
                take_action(action, args, option_string)
            return stop

        # the list of Positionals left to be parsed; this is modified
        # by consume_positionals()
        positionals = self._get_positional_actions()

        # function to convert arg_strings into positional actions
        def consume_positionals(start_index):
            # match as many Positionals as possible
            match_partial = self._match_arguments_partial
            selected_pattern = arg_strings_pattern[start_index:]
            arg_counts = match_partial(positionals, selected_pattern)

            # slice off the appropriate arg strings for each Positional
            # and add the Positional and its args to the list
            for action, arg_count in zip(positionals, arg_counts):
                args = arg_strings[start_index: start_index + arg_count]
                start_index += arg_count
                take_action(action, args)

            # slice off the Positionals that we just parsed and return the
            # index at which the Positionals' string args stopped
            positionals[:] = positionals[len(arg_counts):]
            return start_index

        # consume Positionals and Optionals alternately, until we have
        # passed the last option string
        extras = []
        start_index = 0
        if option_string_indices:
            max_option_string_index = max(option_string_indices)
        else:
            max_option_string_index = -1
        while start_index <= max_option_string_index:

            # consume any Positionals preceding the next option
            next_option_string_index = min([
                index
                for index in option_string_indices
                if index >= start_index])
            if start_index != next_option_string_index:
                positionals_end_index = consume_positionals(start_index)

                # only try to parse the next optional if we didn't consume
                # the option string during the positionals parsing
                if positionals_end_index > start_index:
                    start_index = positionals_end_index
                    continue
                else:
                    start_index = positionals_end_index

            # if we consumed all the positionals we could and we're not
            # at the index of an option string, there were extra arguments
            if start_index not in option_string_indices:
                strings = arg_strings[start_index:next_option_string_index]
                extras.extend(strings)
                start_index = next_option_string_index

            # consume the next optional and any arguments for it
            start_index = consume_optional(start_index)

        # consume any positionals following the last Optional
        stop_index = consume_positionals(start_index)

        # if we didn't consume all the argument strings, there were extras
        extras.extend(arg_strings[stop_index:])

        # if we didn't use all the Positional objects, there were too few
        # arg strings supplied.
        if positionals:
            self.error(_('too few arguments'))

        # make sure all required actions were present
        for action in self._actions:
            if action.required:
                if action not in seen_actions:
                    name = argparse._get_action_name(action)
                    self.error(_('argument %s is required') % name)

        # make sure all required groups had one option present
        for group in self._mutually_exclusive_groups:
            if group.required:
                for action in group._group_actions:
                    if action in seen_non_default_actions:
                        break

                # if no actions were used, report the error
                else:
                    names = [argparse._get_action_name(action)
                             for action in group._group_actions
                             if action.help is not argparse.SUPPRESS]
                    msg = _('one of the arguments %s is required')
                    self.error(msg % ' '.join(names))

        # return the updated namespace and the extra arguments
        return namespace, extras

    def _match_argument(self, action, arg_strings_pattern):
        # match the pattern for this action to the arg strings
        nargs_pattern = self._get_nargs_pattern(action)
        match = argparse._re.match(nargs_pattern, arg_strings_pattern)

        # raise an exception if we weren't able to find a match
        if match is None:
            nargs_errors = {
                None: _('expected one argument'),
                argparse.OPTIONAL: _('expected at most one argument'),
                argparse.ONE_OR_MORE: _('expected at least one argument'),
            }
            default = _('expected %s argument(s)') % action.nargs
            msg = nargs_errors.get(action.nargs, default)
            raise argparse.ArgumentError(action, msg)

        # return the number of arguments matched
        return len(match.group(1))

    def _parse_optional(self, arg_string):
        # if it's an empty string, it was meant to be a positional
        if not arg_string:
            return None

        # if it doesn't start with a prefix, it was meant to be positional
        if not arg_string[0] in self.prefix_chars:
            return None

        # if the option string is present in the parser, return the action
        if arg_string in self._option_string_actions:
            action = self._option_string_actions[arg_string]
            return action, arg_string, None

        # if it's just a single character, it was meant to be positional
        if len(arg_string) == 1:
            return None

        # if the option string before the "=" is present, return the action
        if '=' in arg_string:
            option_string, explicit_arg = arg_string.split('=', 1)
            if option_string in self._option_string_actions:
                action = self._option_string_actions[option_string]
                return action, option_string, explicit_arg

        # search through all possible prefixes of the option string
        # and all actions in the parser for possible interpretations
        option_tuples = self._get_option_tuples(arg_string)

        # if multiple actions match, the option string was ambiguous
        if len(option_tuples) > 1:
            options = ', '.join([option_string
                for action, option_string, explicit_arg in option_tuples])
            tup = arg_string, options
            self.error(_('ambiguous option: %s could match %s') % tup)

        # if exactly one action matched, this segmentation is good,
        # so return the parsed action
        elif len(option_tuples) == 1:
            option_tuple, = option_tuples
            return option_tuple

        # if it was not found as an option, but it looks like a negative
        # number, it was meant to be positional
        # unless there are negative-number-like options
        if self._negative_number_matcher.match(arg_string):
            if not self._has_negative_number_optionals:
                return None

        # if it contains a space, it was meant to be a positional
        if ' ' in arg_string:
            return None

        # it was meant to be an optional but there is no such option
        # in this parser (though it might be a valid option in a subparser)
        return None, arg_string, None

    def _get_option_tuples(self, option_string):
        result = []

        # option strings starting with two prefix characters are only
        # split at the '='
        chars = self.prefix_chars
        if option_string[0] in chars and option_string[1] in chars:
            if '=' in option_string:
                option_prefix, explicit_arg = option_string.split('=', 1)
            else:
                option_prefix = option_string
                explicit_arg = None
            for option_string in self._option_string_actions:
                if option_string.startswith(option_prefix):
                    action = self._option_string_actions[option_string]
                    tup = action, option_string, explicit_arg
                    result.append(tup)

        # single character options can be concatenated with their arguments
        # but multiple character options always have to have their argument
        # separate
        elif option_string[0] in chars and option_string[1] not in chars:
            option_prefix = option_string
            explicit_arg = None
            short_option_prefix = option_string[:2]
            short_explicit_arg = option_string[2:]

            for option_string in self._option_string_actions:
                if option_string == short_option_prefix:
                    action = self._option_string_actions[option_string]
                    tup = action, option_string, short_explicit_arg
                    result.append(tup)
                elif option_string.startswith(option_prefix):
                    action = self._option_string_actions[option_string]
                    tup = action, option_string, explicit_arg
                    result.append(tup)

        # shouldn't ever get here
        else:
            self.error(_('unexpected option string: %s') % option_string)

        # return the collected option tuples
        return result


    # ========================
    # Value conversion methods
    # ========================

    def _get_value(self, action, arg_string):
        type_func = self._registry_get('type', action.type, action.type)
        if not _callable(type_func):
            msg = _('%r is not callable')
            raise argparse.ArgumentError(action, msg % type_func)

        # convert the value to the appropriate type
        try:
            result = type_func(arg_string)

        # ArgumentTypeErrors indicate errors
        except argparse.ArgumentTypeError:
            name = getattr(action.type, '__name__', repr(action.type))
            msg = str(argparse._sys.exc_info()[1])
            raise argparse.ArgumentError(action, msg)

        # TypeErrors or ValueErrors also indicate errors
        except (TypeError, ValueError):
            name = getattr(action.type, '__name__', repr(action.type))
            msg = _('invalid %s value: %r')
            raise argparse.ArgumentError(action, msg % (name, arg_string))

        # return the converted value
        return result

    def _check_value(self, action, value):
        # converted value must be one of the choices (if specified)
        if action.choices is not None and value not in action.choices:
            tup = value, ', '.join(map(repr, action.choices))
            msg = _('invalid choice: %r (choose from %s)') % tup
            raise argparse.ArgumentError(action, msg)

    def error(self, message):
        """error(message: string)

        Prints a usage message incorporating the message to stderr and
        exits.

        If you override this in a subclass, it should not return -- it
        should either exit or raise an exception.
        """
        self.print_usage(argparse._sys.stderr)
        self.exit(2, (_('%s: error: %s') + '\n') % (self.prog, message))



"""Classes that communicate the results."""

import sys
from itertools import zip_longest
from collections import defaultdict
import pvcheck.executor
from pvcheck.i18n import translate as _


def handle_non_printable_chars(string):
    if string.isprintable():
        out_string = string
    else:
        out_string = ''
        for c in string:
            if c == '\n' or c.isprintable():
                out_string = ''.join([out_string, c])
            else:
                out_string = ''.join([out_string, '<NP>'])
    return out_string


class Formatter:
    """Abstract base class for all the formatters.

    The client is supposed to call the methods in the following order:

    1) begin_session
    2) <repeat zero or more times>
    2a)  begin_test
    2b)  execution_result
    2c)  zero or more of comparison_result or missing_section
    2d)  end_test
    3) end_session

    """

    def begin_session(self):
        """Called when starting a new test session."""
        pass

    def end_session(self):
        """Called at the end of a test session."""
        pass

    def begin_test(self, description, cmdline_args, input, tempfile):
        """Called when a new test begins."""
        pass

    def end_test(self):
        """Called when a test finishes."""
        pass

    def execution_result(self, cmdline_args, execution_result, test):
        """Called when the execution of a test is terminated."""
        pass

    def comparison_result(self, expected, got, diffs, matches):
        """Called after the comparison of a section.

        - expected: the expected result
        - got: the output of the program
        - diffs: distance measures between the two sets of lines
        - matches: lines matching those in 'got'
        """
        pass

    def missing_section(self, expected):
        """Called when no answer has been provided for a section.

        - expected: the expected result
        """
        pass


class TextFormatter(Formatter):
    """Formatter that writes the results as plain text."""

    # Message priority levels, corresponding to the verbosity.
    DEBUG = 4
    INFO = 3
    SUCCESS = 2
    WARNING = 1
    ERROR = 0
    FATAL = -1

    _RESULT_TABLE = {
        pvcheck.executor.ER_OK: (DEBUG, []),
        pvcheck.executor.ER_TIMEOUT:
        (ERROR, [_("TIMEOUT EXPIRED: PROCESS TERMINATED")]),
        pvcheck.executor.ER_OUTPUT_LIMIT:
        (ERROR, [_("TOO MANY OUTPUT LINES")]),
        pvcheck.executor.ER_SEGFAULT:
        (ERROR, [_("PROCESS ENDED WITH A FAILURE"),
                 _("(SEGMENTATION FAULT)")]),
        pvcheck.executor.ER_ERROR:
        (ERROR, [_("PROCESS ENDED WITH A FAILURE"),
                 _("(ERROR CODE {status})")]),
        pvcheck.executor.ER_NOTFILE:
        (ERROR, [_("FAILED TO RUN THE FILE '{progname}'"),
                 _("(the file does not exist)")])
    }

    def __init__(self, destination=sys.stdout, verbosity=None,
                 maxerrors=None):
        """Create the text formatter.

        - destination: file-like object receiving the text messages
        - verbosity: set the amount of information to write
        - maxerrors: maximum number of errors to print (per section)

        verbosity must be in the range 0-4.  maxerrors must be at
        least 1 (or None to print an unlimited number of errors).

        """
        self.set_verbosity(verbosity)
        self._dst = destination
        self._maxerrors = maxerrors
        self._testcount = 0
        self._test_status = None

    def set_verbosity(self, verbosity=None):
        """Set a new verbosity level (0-4)."""
        self._verbosity = (verbosity if verbosity is not None
                           else self.INFO)

    def message(self, level, text):
        """Write a text message with the given level."""
        if self.level_enabled(level):
            self._dst.write(text)
            self._dst.write('\n')

    def level_enabled(self, level):
        """Return true if the output level is enabled."""
        return (self._verbosity >= level)

    def debug(self, text):
        """Write a message with the debug level."""
        self.message(self.DEBUG, text)

    def info(self, text):
        """Write a message with the info level."""
        self.message(self.INFO, text)

    def success(self, text):
        """Write a message with the success level."""
        self.message(self.SUCCESS, text)

    def warning(self, text):
        """Write a message with the warning level."""
        self.message(self.WARNING, text)

    def error(self, text):
        """Write a message with the error level."""
        self.message(self.ERROR, text)

    def fatal(self, text):
        """Write a message with the fatal level."""
        self.message(self.FATAL, text)

    def _format_section(self, title, content, maxlines=5):
        """Compose a text session.

        If there are too many lines the string is truncated with a
        message.

        """
        lines = list(map(str.rstrip, content.splitlines()))
        n = len(lines)
        if n == 0:
            return "{}: <empty>".format(title)
        if n == 1:
            return "{}: {}".format(title, lines[0])
        if maxlines is not None and n > maxlines:
            lines = lines[:maxlines]
            extra = n - maxlines + 1
            lines[-1] = _("(... plus other %d lines ...)") % extra
        return "{}:\n{}".format(title, "\n".join(lines))

    def begin_session(self):
        # Initialize the counters for the summary
        self._testcount = 0
        self._sect_results = []

    def _unique(self, seq):
        # Unique elements in seq, preserving their order of occurrence
        seen = set()
        for x in seq:
            if x in seen:
                continue
            seen.add(x)
            yield x

    def _count_results(self, tag, label):
        return sum(1 for x in self._sect_results if x == (tag, label))

    def end_session(self):
        if self._testcount < 2:
            return

        # Write a summary of the session
        self.info("")
        self.info("=" * 60)
        self.info("")
        self.info(_("SUMMARY"))

        tags = list(self._unique(x[0] for x in self._sect_results))
        l = max(map(len, tags))
        for t in tags:
            row = ["{:{}}:".format(t, l)]

            row.append("%2d %s," % (self._count_results(t, "ok"), _("successes")))
            row.append("%2d %s," % (self._count_results(t, "warning"), _("warnings")))
            row.append("%2d %s" % (self._count_results(t, "error"), _("errors")))
            self.info("  ".join(row))
        self.info("")
        self._testcount = 0

    def _proc_args(self, args):
        return [(a if a is not pvcheck.executor.ARG_TMPFILE
                 else _("<temp.file>"))
                for a in args]

    def begin_test(self, description, cmdline_args, input, tempfile):
        if self._testcount > 0:
            self.info("-" * 60)
        self._testcount += 1
        maxlines = (None if self.level_enabled(self.DEBUG) else 5)
        f = lambda tit,con: self._format_section(tit, con, maxlines)
        if description is not None:
            self.info(f(_("TEST"), description))
        arglist = " ".join(self._proc_args(cmdline_args))
        self.info(f(_("COMMAND LINE"), arglist))
        if input is not None and input.strip():
            self.info(f(_("INPUT"), input))
        if tempfile is not None:
            self.info(f(_("TEMPORARY FILE"), tempfile))
        self._test_status = None  # initialize test status

    def end_test(self):
        if self._test_status == "ok":
            self._sect_results.append((_("<program>"), "ok"))
        elif self._test_status == "warning":
            self._sect_results.append((_("<program>"), "warning"))
        else:
            self._sect_results.append((_("<program>"), "error"))

    def execution_result(self, cmdline_args, execution_result, test):
        info = {
            'progname': cmdline_args[0],
            'status': execution_result.status
        }
        level, lines = self._RESULT_TABLE[execution_result.result]
        msg = " ".join(lines).format(**info)
        if msg:
            self.message(level, msg)
        if execution_result.output and self.level_enabled(self.DEBUG):
            # This can be a lot of data, so do it only when the DEBUG
            # level is enabled.
            lines = "\n".join("> " + line for line
                     in execution_result.output.splitlines())
            self.debug(self._format_section(_("OUTPUT"),
                                            lines,
                                            maxlines=None))

    def comparison_result(self, expected, got, diffs, matches):
        if max(diffs, default=0) == 0:
            self._sect_results.append((expected.tag, "ok"))
            if self._test_status is None:
                self._test_status = "ok"
            self.success("{}: {}".format(expected.tag, _("OK")))
        else:
            self._sect_results.append((expected.tag, "error"))
            self._test_status = "error"
            if len(expected.content) != len(got.content):
                fmt = _("wrong number of lines (expected %d, got %d)")
                msg = fmt % (len(expected.content), len(got.content))
                self.error(expected.tag + ": " + msg)

            err_diff = _("line %d is wrong  (expected '%s', got '%s')")
            err_un = _("unexpected line '%s'")
            err_mis = _("missing line (expected '%s')")
            lines = []
            for (i, d) in enumerate(diffs):
                if d <= 0:
                    continue
                try:
                    out_string = handle_non_printable_chars(got.content[i])
                except IndexError:
                    out_string = ''
                if matches[i] is None:
                    msg = err_un % out_string
                elif i >= len(got.content):
                    msg = err_mis % matches[i]
                else:
                    msg = err_diff % (i + 1, matches[i], out_string)
                lines.append(expected.tag + ": " + msg)

            disp = (len(lines) if self._maxerrors is None
                    else (len(lines) if len(lines) < self._maxerrors
                          else self._maxerrors - 1))
            extra = len(lines) - disp
            for line in lines[:disp]:
                self.error(line)
            if extra > 0:
                self.error(_("(... plus other %d errors ...)") % extra)
            if len(got.content) > 0 and max(diffs[:len(got.content)]) == 0:
                fmt = _("The first %d lines matched correctly")
                msg = fmt % len(got.content)
                self.warning(expected.tag + ": " + msg)

        # Extra information to be skipped unless the DEBUG level is
        # enabled.
        if self.level_enabled(self.DEBUG):
            self._detailed_comparison(expected, got, matches)

    def _detailed_comparison(self, expected, got, matches):
        fmt = "%-30s| %-30s"
        def prnt(s):
            s = (s.rstrip() if s is not None else _("<nothing>"))
            return (s if len(s) <= 30 else s[:26] + "...")

        self.debug(expected.tag + ": " +  _("detailed comparison"))
        self.debug(fmt % (_("EXPECTED OUTPUT"), _("ACTUAL OUTPUT")))

        for (e, a) in zip_longest(matches, got.content):
            self.debug(fmt % (prnt(e), prnt(a)))

    def missing_section(self, expected):
        if self._test_status != "ok" or self._test_status is None:
            self._test_status = "warning"
        self._sect_results.append((expected.tag, "warning"))
        self.warning(expected.tag + ": " + _("missing section"))


class ColoredTextFormatter(TextFormatter):
    """A formatter that outputs colored messages.

    The color is chosen according to the type of the message.
    """

    BLUE = '\033[94m'
    DARK_BLUE = '\033[34m'
    YELLOW = '\033[93m'
    DARK_YELLOW = '\033[33m'
    GREEN = '\033[92m'
    DARK_GREEN = '\033[92m'
    RED = '\033[91m'
    DARK_RED = '\033[91m'
    GRAY = '\033[97m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'

    COLORS = {
        TextFormatter.FATAL: BOLD + DARK_RED,
        TextFormatter.ERROR: BOLD + RED,
        TextFormatter.WARNING: BOLD + DARK_YELLOW,
        TextFormatter.INFO: "",
        TextFormatter.SUCCESS: BOLD + DARK_GREEN,
        TextFormatter.DEBUG: BLUE
    }

    def message(self, level, text):
        super().message(level, self.COLORS[level] + text + self.ENDC)


class CombinedFormatter(Formatter):
    """A formatter that broadcasts the messages to other formatters."""

    def __init__(self, formatters=()):
        """Create the formatter."""
        self.formatters = list(formatters)

    def begin_session(self):
        for f in self.formatters:
            f.begin_session()

    def end_session(self):
        for f in self.formatters:
            f.end_session()

    def begin_test(self, *args):
        for f in self.formatters:
            f.begin_test(*args)

    def end_test(self):
        for f in self.formatters:
            f.end_test()

    def execution_result(self, *args):
        for f in self.formatters:
            f.execution_result(*args)

    def comparison_result(self, *args):
        for f in self.formatters:
            f.comparison_result(*args)

    def missing_section(self, *args):
        for f in self.formatters:
            f.missing_section(*args)



# -*- coding: utf-8 -*-

"""Internationalization.

Use

  from i18n import translate as _

to enable string localization. In the future it would be easily
replaced by gettext since translatable strings are marked in the
same way as _("text").

Beside the translation service, this module defines the messages
used in the application, such as the help and the usage messages.
"""

import os
import sys
import atexit


############################################################
# Translation function
############################################################

def translate(text):
    """Translate the text in the current language."""
    try:
        return _strings[text]
    except KeyError:
        pass
    # try removing the last char since translation loading strips the end spaces
    try:
        return _strings[text[:-1]] + text[-1]
    except KeyError:
        _missing_translation.add(text)
        s = (text if len(text) < 15 else text[:15] + "...")
        print("[TT] Warning: missing translation for '%s'" % s, file=sys.stderr)
        return text


############################################################
# Extra functions
############################################################

def register_translation(english_text, language, translation):
    """Function to add translations from other modules."""
    if _lang == language:
        _strings[english_text] = translation
    elif _lang == "en":
        _strings[english_text] = english_text


############################################################
# Application messages
############################################################

_strings = {}
_missing_translation = set()
_lang = 'en'

_translations = {
    'it': {
        "Run tests to verify the correctness of a program.": "Esegue dei test per verificare la correttezza di un programma.",
        "set the verbosity level, where the level must be an integer between 0 (minimum) and 4 (maximum). The default value is 3.": "imposto il livello di verbosità.  Il livello deve essere un valore intero tra 0 (minimo) e 3 (massimo).  Il default è 2.",
        "set how many seconds it should be waited for the termination of the program.  The default is 10 seconds.": "imposta per quanti secondi bisogna attendere la terminazione del programma.  Il default è pari a 10 secondi.",
        "cut the output of the program to a maximum of L lines.  The default is 10000.": "taglia l'output del programma ad un massimo di L linee.  Il default è 10000.",
        "reports up to N errors per section (default 4).": "riporta fino ad un massimo di N errori per sezione (default 4).",
        "uses the specified configuration file.": "utilizza il file di configurazione specificato.",
        "enable or disable colored output (default AUTO).": "abilita o disabilita l'output colorato (default AUTO).",
        "use Valgrind (if installed) to check memory usage.": "utilizza Valgrind (se installato) per controllare l'utilizzo della memoria.",
        "select the output type.": "seleziona il tipo di output.",
        "specify the name of the file used for logging.  The default is ~/.pvcheck.log.": "specifica il nome del file usato per il logging. Il default è ~/.pvcheck.log.",
        "print this message and exit.": "stampa questo messaggio ed esce.",
        "list all the available tests.": "mostra tutti i test disponibili.",
        "run only the selected test.": "esegue solo il test indicato.",
        "export in a file the input arguments from the selected test.": "salva in un file gli argomenti di input dal test indicato.",
        "ERROR": "ERRORE",
        "OK": "OK",
        "WARNING!": "ATTENZIONE!",
        "COMMAND LINE": "RIGA DI COMANDO",
        "<temp.file>": "<file temp.>",
        "INPUT": "INPUT",
        "Input:": "Input:",
        "OUTPUT": "OUTPUT",
        "TEMPORARY FILE": "FILE TEMPORANEO",
        "TEST": "TEST",
        "line %d": " riga %d",
        "(expected '%s'": " (atteso '%s'",
        ", got '%s')": "ottenuto '%s')",
        "unexpected line '%s'": "riga inattesa '%s'",
        "missing line (expected '%s')": "riga mancante (atteso '%s')",
        "wrong number of lines (expected %d, got %d)": "numero di righe errato (atteso %d, ottenuto %d)",
        "line %d is wrong  (expected '%s', got '%s')": "riga %d errata  (atteso '%s', ottenuto '%s')",
        "The first %d lines matched correctly": "Le prime %d righe sono corrette",
        "(... plus other %d errors ...)": "(... più altri %d errori ...)",
        "ACTUAL OUTPUT": "OUTPUT EFFETTIVO",
        "EXPECTED OUTPUT": "OUTPUT ATTESO",
        "detailed comparison": "confronto dettagliato",
        "<nothing>": "<niente>",
        "missing section": "sezione mancante",
        "empty section": "sezione vuota",
        "extra section": "sezione extra",
        "Invalid parameter": "Parametro non valido",
        "Invalid parameter ('%s')": "Parametro non valido ('%s')",
        "Invalid parameter ('%d')": "Parametro non valido('%d')",
        "Invalid parameter ('%f')": "Parametro non valido('%f')",
        "TIMEOUT EXPIRED: PROCESS TERMINATED": "TEMPO LIMITE SCADUTO: PROCESSO TERMINATO",
        "TOO MANY OUTPUT LINES": "TROPPE LINEE DI OUTPUT",
        "PROCESS ENDED WITH A FAILURE": "PROCESSO TERMINATO CON UN FALLIMENTO",
        "(SEGMENTATION FAULT)": "(SEGMENTATION FAULT)",
        "(ERROR CODE {status})": "(CODICE D'ERRORE {status})",
        "FAILED TO RUN THE FILE '{progname}'": "IMPOSSIBILE ESEGUIRE IL FILE '{progname}'",
        "(the file does not exist)": "(file inesistente)",
        "(... plus other %d lines ...)": "(... più altre %d righe ...)",
        "SUMMARY": "RIEPILOGO",
        "Summary": "Riepilogo",
        "summary": "riepilogo",
        "successes": "successi",
        "Successes": "Successi",
        "warnings": "avvertimenti",
        "Warnings": "Avvertimenti",
        "errors": "errori",
        "Errors": "Errori",
        "<program>": "<programma>",
        "CODE": "CODICE",
        "TOTAL": "TOTALE",
        "Test number %d doesn't exist.": "Il test numero %d non esiste.",
        "Use './pvcheck info' to list all the available tests.": "Utilizza './pvcheck info' per vedere tutti i test disponibili.",
        "Error: Can't export test number %d.": "Errore: Impossibile esportare il test numero %d.",
        "file containing the tests to be performed (default pvcheck.test).": "file contenente i test da eseguire (default pvcheck.test).",
        "file containing the tests to be performed.": "file contenente i test da eseguire.",
        "program to be tested.": "programma da testare.",
        "any arguments of the program to be tested.": "eventuali argomenti del programma da testare.",
        "[run|info|export] --help for command help (default=run)": "[run|info|export] --help per l'help di un comando (default=run)",
        "test a program.": "testa un programma.",
        "Test Result": "Risultato Test",
        "positional arguments": "argomenti posizionali",
        "optional arguments": "argomenti opzionali",
        "show this help message and exit": "mostra questo messaggio ed esce",
        "show program's version number and exit": "mostra la versione del programma ed esce",
        "unrecognized arguments: %s'": "argomento %s non riconosciuto",
        "not allowed with argument %s": "l'argomento %s non é consentito",
        "ignored explicit argument %r": "ignorato l'argomento esplicito %r",
        "too few arguments": "troppi pochi argomenti ",
        "argument %s is required": "é necessario l'argomento %s",
        "one of the arguments %s is required": "é necessario uno dei seguenti argomenti %s",
        "expected one argument": "atteso un argomento",
        "expected at most one argument": "atteso al piú un argomento",
        "expected at least one argument": "atteso almeno un argomento",
        "expected %s argument(s)": "atteso argomento %s",
        "ambiguous option: %s could match %s": "opzione ambigua: %s puó coincidere con %s",
        "unexpected option string: %s": "opzione string non attesa: %s",
        "%r is not callable": "%r non é chiamabile",
        "invalid %s value: %r": "non valido %s valore: %r",
        "invalid choice: %r (choose from %s)": "scelta non valida: %r (i parametri disponibili sono %s)",
        "%s: error: %s": "%s: errore: %s",
        "unrecognized arguments: %s": "argomento non riconosciuto: %s",
        "Command line: %s": "Riga di comando: %s",
        "FAILED TO RUN THE FILE '{progname}' the file does not exist)": "ERRORE NELL'ESECUZIONE DEL FILE '{progname}' il file non esiste)",
        "Lines %d-%d/%d": "Righe %d-%d/%d",
        "PROCESS ENDED WITH A FAILURE (SEGMENTATION FAULT)": "IL PROCESSO E` TERMINATO CON UN FALLIMENTO (SEGMENTATION FAULT)",
        "PROGRAM'S OUTPUT:": "OUTPUT DEL PROGRAMMA:",
        "SUMMARY:": "SOMMARIO:",
        "TEMP_FILE": "FILE_TEMPORANEO",
        "Temporary file:": "File temporaneo:",
        "Test case %d of %d (%s)": "Caso di test %d di %d (%s) ",
        "Test title: %s": "Titolo del test: %s",
        "[Press 'h' for help]": "[Premere 'h' per l'aiuto]",
        "enables the interactive mode.": "abilita la modalita` interattiva.",
        "expected": "atteso",
        "missing line": "riga mancante",
        "passes": "passati",
        "this line was not expected": "questa riga e` inattesa",
        "TEST RUNNING": "TEST IN ESECUZIONE",
        "TEST COMPLETED": "TEST TERMINATO",
        "section [%s] is missing": "sezione [%s] non trovata",
        "execution failed": "esecuzione fallita",
        "%(prog)s: error: %(message)s\n": "%(prog)s: errore: %(message)s\n",
        "the following arguments are required: %s": "i seguenti argomenti sono richiesti: %s",
        "usage: ": "utilizzo: ",
        'file containing the tests to be exported.': "file contenente i test da salvare.",
        "number of the test to export as returned by the 'info' command.": "numero del test da salvare come restituito dal comando 'info'."
    }
}


############################################################
# Setup and shutdown
############################################################

def _install_lang():
    global _lang
    global _strings
    _lang = os.environ.get('LANG', 'en').partition('_')[0]
    try:
        _strings = _translations[_lang]
    except KeyError:
        # Fallback to english, where the translation is the identity
        # function.
        _lang = 'en'
        for t in _translations.values():
            _strings = {k: k for k in t}
            break


_install_lang()


@atexit.register
def _save_missing_translations():
    if _missing_translation:
        import pprint
        with open('missing_translations.txt', 'wt') as f:
            d = dict.fromkeys(_missing_translation, "")
            pprint.pprint({_lang: d}, stream=f)



"""Matching functions.

Used to compare the expected and the actual outputs.
"""

import re
from itertools import zip_longest


_RE_INT = re.compile(r'(-|\+)?[0-9]+')
_RE_REAL = re.compile(r'(-|\+)?[0-9]+\.(?P<frac>[0-9]*)')


def compare_sections(actual, expected, ordered=True):
    """Compare the content of two sections.

    The 'ordered' parameter indicates whether or not the order of the
    expected lines matter.  The function returns a pair of lists: a
    list of differences and a list of matches.

    The list of difference contains difference measures between given
    and expected lines of text (with 0 being a perfect match).

    The list of matches contains elements from the 'expected' section,
    in the order in which they have been matched to the elements of
    the 'actual' section.  The lists may contain 'None' values when
    actual lines have not been successfully matched.  The list cannot
    be shorter than the first parameter, but may also be longer when
    expected lines are not matched.  When 'ordered' is True this list
    is a copy of the elements in 'expected', possibly with trailing
    Nones.

    In practice a call to zip_longest with the 'actual' lines and the
    returned matches would produce a side-by-side comparison of the
    matched lines, with None values filling the holes.

    """
    if ordered:
        matched = list(expected)
        diffs = _compare_ordered_sections(actual, matched)
        matched.extend([None] * (len(actual) - len(matched)))
    else:
        diffs, matched = _compare_unordered_sections(actual, expected)
    return (diffs, matched)


def _compare_elements(value, expected):
    """Compare a pair of elements.

    Return true if value matches expected value.  The type of
    comparison (numerical with a set precision or textual) is inferred
    from the expected value.

    """
    m = _RE_REAL.fullmatch(expected)
    if m is not None:
        # If the expected value is a fractional number, compare the
        # result up to the given number of digits.
        digits = len(m.group('frac'))
        try:
            return float(expected) == round(float(value), digits)
        except ValueError:
            return False
    elif _RE_INT.fullmatch(expected):
        # If the expected value is an integer, compare by value so
        # that 15, +15 and 015 are considered as equivalent.
        try:
            return int(expected) == int(value)
        except ValueError:
            return False
    else:
        # In the general case just compare the strings.
        return (value == expected)


def _compare_lines(actual, expected):
    """Return the difference between the given and the expected lines.

    The result is in the range [0, 1], with 0 denoting a perfect
    match.

    """
    act = actual.split()
    exp = expected.split()
    ok = sum(1 for a, e in zip(act, exp) if _compare_elements(a, e))
    den = max(len(act), len(exp))
    return float(den - ok) / max(den, 1)


def _compare_ordered_sections(actual, expected):
    """Compare the content of two sections.

    The result is a list of distance measures between pairs of actual
    and expected lines.  The length of the list is the same of the
    longest of the two sections.

    """
    return [1.0 if x[0] is None or x[1] is None
            else _compare_lines(x[0], x[1])
            for (i, x) in enumerate(zip_longest(actual, expected))]


def _compare_unordered_sections(actual, expected):
    """Compare two sections without considering the order of their lines.

    The result is a pair of lists: the first list contains the
    distance measures between pairs of actual and expected lines.  The
    second list contains the expected lines reordered to match the
    actual ones, possibily with the inclusion of additional None
    values where no match is found.  The tow lists have the same
    length.

    """
    expected = set(enumerate(expected))
    diffs = []
    ordered = []
    for x in actual:
        for (n, y) in expected:
            d = _compare_lines(x, y)
            if d == 0:
                diffs.append(d)
                ordered.append(y)
                expected.remove((n, y))
                break
        else:
            diffs.append(1.0)
            ordered.append(None)
    for (n, y) in sorted(expected):
        diffs.append(1.0)
        ordered.append(y)
    return (diffs, ordered)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pvcheck']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['pvcheck = pvcheck.main:main']}

setup_kwargs = {
    'name': 'pvcheck',
    'version': '0.1.6',
    'description': 'Perform automatic tests of the output produced by command line programs',
    'long_description': 'pvcheck\n=======\n\npvcheck is a tool that automates the testing of computer programs by checking the textual output of the program under testing with a set of user-defined test cases.\n\npvcheck has been developed to be used as automatic correction tool in the courses of Principles of Computer Programming at the University of Pavia.\n\nRequirements\n-------------\n\npvcheck requires a Python interpreter and can be executed as a standalone application.\nThe current version has been tested with Python 3.4.3.\n\nInstallation\n------------\n\nTo install pvcheck:\n\n```\npip install pvcheck\n```\n\nTo compile all the sample programs:\n\n```\ncd examples\nmake\n```\n\nUsage\n-----\n\npvcheck has 3 main arguments ([run](#run), [info](#info) and [export](#export)) .\n\n### help ###\n\nTo see the main help:\n\n```\npvcheck -h\n```\n\nor \n\n```\npvcheck --help\n```\n\nTo see the help of a specific command:\n\n```\npvcheck argument -h\n```\n\nor\n\n```\npvcheck argument --help\n```\n\n### the <a name="run"></a>run argument ###\n\nThe run argument, being the default argument, also works without explicit invocation. So\n\n```\npvcheck run ./program\n```\n\nis the same as:\n\n```\npvcheck ./program\n```\n\n#### test a program with a default test file ####\n\nIf you have a [test file](#testfile) with the default name *pvcheck.test* located in the pvcheck folder, you can test your program simply by using:\n\n```\npvcheck ./program\n```\n\n#### test a program with any test file ####\n\nIf you want to test a program using a specific [test file](#testfile):\n\n```\npvcheck -f testfile ./program\n```\n```\npvcheck --file testfile ./program\n```\n\n#### test a program with a specific test of a test suite ####\n\nIf you want to test a program using a specific test contained in a test suite, you can use the [info argument](#info) to list all the available tests. Then you can use:\n\n```\npvcheck -T testnumber ./program\n``` \n```\npvcheck --test testnumber ./program\n``` \n\nto use only the desired test.\n\n#### select a different output format ####\n\npvcheck lets you choose the output format. The formats available are [json](https://github.com/claudio-unipv/pvcheck/wiki/JSON-data-format), csv and html. To choose the ouput format:\n\n```\npvcheck -F format ./program\n``` \n```\npvcheck --format format ./program\n``` \n\n#### change timeout ####\n\nYou can set how many seconds it should be waited for the termination of the program:\n\n```\npvcheck -t timeout ./program\n```\n```\npvcheck --timeout timeout ./program\n``` \n\nThe default is 10 seconds.\n\n#### change the number of reported errors #####\n\nTo report up to N errors per section:\n\n```\npvcheck -e N ./program\n``` \n```\npvcheck --errors N ./program\n``` \n\nThe default is 4.\n\n#### set the verbosity level ####\n\nTo set the verbosity level, where the level must be an integer between 0 (minimum) and 4 (maximum):\n\n```\npvcheck -v verbositylevel ./program\n``` \n```\npvcheck --verbosity verbositylevel ./program\n``` \n\nThe default is 3.\n\n#### use valgrind ####\n\nTo use Valgrind (if installed) to check memory usage:\n\n```\npvcheck -V ./program\n``` \n```\npvcheck --valgrind ./program\n``` \n\n#### use a log file ####\n\nTo specify the name of the file used for logging:\n\n```\npvcheck -l logfile ./program\n``` \n```\npvcheck --log logfile ./program\n``` \n\nThe default is ~/.pvcheck.log.\n\n#### set an output limit ####\n\nTo cut the output of the program to a maximum of L lines:\n\n```\npvcheck -L L ./program\n``` \n```\npvcheck --output_limit L ./program\n```\n\nThe default is 10000.\n\n#### use a configuration file ####\n\nTo use the specified configuration file:\n\n\n```\npvcheck -c configurationfile ./program\n``` \n```\npvcheck --config configurationfile ./program\n``` \n\n#### change output color setting ####\n\nTo enable or disable colored output:\n\n```\npvcheck -C option ./program\n```\n```\npvcheck --color option ./program\n```\n\nThe options available are *YES*, *NO* and *AUTO*. The default value is *AUTO*.\n\n### the <a name="info"></a>info argument ###\n\nThe info argument lists all the available tests in a test file. \n\nTo see all the tests available in a test file:\n\n```\npvcheck info testfile\n``` \n\n### the <a name="export"></a>export argument ###\n\nThe export argument allows you to export in a file the input arguments from the selected test.\n\nTo export the input argument from a test, you can use the [info argument](#info) to list all the available tests. Then you can use:\n\n```\npvcheck export testnumber testfile\n``` \n\nThe output file is saved into the current directory with the name testname.dat .\n\nTest File<a name="testfile"></a>\n---------\n\n### description ###\n\nA test file is a file which contains one or more test cases.\n\nTests are divided in sections with names between square brackets.\n\nEmpty lines, and lines starting with \'#\' are ignored.\n\nNon-empty lines are compared against those produced by the program in the corresponding section.\n\nThe format of test definition files is rather simple.\nThe following examples include comments to concisely describe the various aspects of the format:\n\n* [single test case](https://github.com/claudio-unipv/pvcheck/blob/master/examples/example.test) \n* [multiple test cases](https://github.com/claudio-unipv/pvcheck/blob/master/examples/example2.test)\n* [multiple output lines](https://github.com/claudio-unipv/pvcheck/blob/master/examples/example3.test)\n* [temporary files](https://github.com/claudio-unipv/pvcheck/blob/master/examples/example4.test)\n\n### special sections ###\n\n#### the special section [.SECTIONS] ####\n\nThe special section [.SECTIONS] allows to specify additional options for the sections.\nOptions are usually declared at the beginning of the test file, or in a separate configuration file.\n\nFor example:\n\n```\n...\n\n[SECTION1]\n...\n\n[.SECTIONS]\nSECTION2 unordered\n\n[SECTION2]\n...\n``` \nindicates that the order of the lines in the SECTION2 section is not relevant.\n\n#### the special section [.TEST] ####\n\nIn case of multiple tests, each test is introduced by the special section [.TEST] followed by the name of the test.\n\nExample:\n\n```\n...\n\n[.TEST]\nTest1\n\n[SECTION1]\n...\n\n[SECTION2]\n...\n\n[.TEST]\nTest2\n\n[SECTION1]\n...\n\n[SECTION2]\n...\n``` \n\nCommon parts among all the test cases (for instance options in the [.SECTIONS] special section) can be specified before the first [.TEST] section. They will be prepended to all the tests.\n\n#### the special section [.INPUT] ####\n\nThe special section [.INPUT] allows to specify the text to be written on the program\'s standard input:\n\nExample:\n\n```\n...\n\n[.TEST]\nName of the test\n\n[.INPUT]\ninput1\ninput2\n...\n\n[SECTION1]\n...\n\n[SECTION2]\n...\n\n[.TEST]\nThi is another test\n\n[.INPUT]\ninput3\ninput4\n...\n\n[SECTION1]\n...\n\n[SECTION2]\n...\n``` \n\n#### the special section [.ARGS] ####\n\nThe special section [.ARGS] allows to specify additional arguments to be passed on the command line, one extra argument per line.\n\nExample:\n\n```\n...\n\n[.TEST]\nTest1\n\n[.ARGS]\narg1\narg2\n...\n\n[SECTION1]\n...\n\n[SECTION2]\n...\n\n[.TEST]\nTest2\n\n[.ARGS]\narg3\narg4\n...\n\n[SECTION1]\n...\n\n[SECTION2]\n...\n``` \n\n#### the special section [.FILE] ####\n\nThe special section [.FILE] shall be used together with the special section [.ARGS].\nWhen the special argument ".FILE" is present in the [.ARGS] section, a temporary file is automatically generated and filled with the content of the text in the [.FILE] section.\nThe name of the temporary file is passed on the command line of the program under test.\n\nExample:\n\n```\n...\n\n[.ARGS]\n.FILE\n...\n\n[.TEST]\nTest1\n\n[.FILE]\nfirst line written in the temporary file used for Test1\nsecond line written in the temporary file used for Test1\n...\n\n[SECTION1]\n...\n\n[SECTION2]\n...\n\n[.TEST]\nTest2\n\n[.FILE]\nfirst line written in the temporary file used for Test2\nsecond line written in the temporary file used for Test2\n...\n\n[SECTION1]\n...\n\n[SECTION2]\n...\n``` \n\nWiki\n----\n\nFor more information, please visit the [wiki](https://github.com/claudio-unipv/pvcheck/wiki) of the project.\n\n',
    'author': 'Claudio Cusano',
    'author_email': 'claudio.cusano@unipv.it',
    'maintainer': 'Claudio Cusano',
    'maintainer_email': 'claudio.cusano@unipv.it',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
